/*************************************************************************
 *
 *  C-to-C Translator
 *
 *  Adapted from Clean ANSI C Parser
 *  Eric A. Brewer, Michael D. Noakes
 *  
 *  $Log: ast.c,v $
 *  Revision 1.14  1996/08/16 23:10:32  randall
 *  Redid how implicit inlets are done.  They are now a type of spawn
 *  statement.  Reorganized the translations of spawn statements a little.
 *
 *  Revision 1.13  1996/06/07 17:55:27  athena
 *  Fixed copyrights. I hate this.
 *
 *  Revision 1.12  1996/06/02 03:20:36  fengmd
 *  Allow pointer arithmetic in implicit inlet.
 *
 *  Revision 1.11  1996/05/21 13:06:20  fengmd
 *  Fix bug on implicit inlet to accumulate structure members, check type mismatch.
 *
 *  Revision 1.10  1996/05/08 20:37:44  randall
 *  Got typedef globalizing right, made slow functions static, a few other
 *  minor adjustments.
 *
 *  Revision 1.9  1996/05/04 01:43:55  randall
 *  Major overhaul.  Made an inlet a type of Proc instead of its own ast
 *  node.  Reorganized dataflow and other files to account for local
 *  procedure nodes.  Changed grammar to allow local function definitions
 *  (including inlets).  Moved most syntax errors into sem-check.c from
 *  transform.c.
 *
 *  Revision 1.8  1996/04/21 04:20:10  fengmd
 *  Add implicit inlet, crlf, on_abort, SYNCHED, SPAWN_COUNTER.
 *
 *  Revision 1.7  1996/03/24 20:08:24  randall
 *  Fixed up ASM directive.
 *
 *  Revision 1.6  1996/03/08 17:07:28  randall
 *  Allowed spawn with void return type in inlet_call.  Started working on
 *  resume and abort.
 *
 *  Revision 1.5  1996/03/05 23:08:35  randall
 *  Added implicit returns and implicit syncs.
 *
 *  Revision 1.4  1996/03/01 21:26:34  randall
 *  Got more inlet stuff working.  Dataflow analysis now works properly on
 *  inlets and inlet_calls.  inlet_call is now translated correctly.
 *
 *  Revision 1.3  1996/02/27 15:48:43  athena
 *  Hacked cilk2c to accept 'asm' (fengmd)
 *
 * Revision 1.2  1996/02/19  22:15:52  randall
 * Moved _PUSH_FRAME/_POP_FRAME calls around, and reorganized rts macros
 * a bit.  Removed #include inlining, and added EnterC and ExitC calls to
 * prevent C headers from being parsed as Cilk.
 *
 *************************************************************************/

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */


#include "ast.h"


Coord UnknownCoord = { 
  /* line:   */ 0,
  /* offset: */ 0,
  /* file:   */	0
  };
			 


/* use HeapNew() (defined in ast.h) to allocate whole objects */
GLOBAL inline void *HeapAllocate(unsigned int number, unsigned int size)
{ 
  return calloc(number, size);
}

GLOBAL inline void HeapFree(void *ptr)
{
  free(ptr);
}



GLOBAL inline Node *NewNode(NodeType typ)
{
    Node *create = HeapNew(Node);

    create->typ = typ;
    create->coord = UnknownCoord;
    create->parenthesized = FALSE;
    create->analysis.livevars = NULL;
    return(create);
}



/*************************************************************************/
/*                                                                       */
/*                          Expression nodes                             */
/*                                                                       */
/*************************************************************************/

GLOBAL inline Node *MakeConstSint(int value)
{ Node *node = NewNode(Const);
  node->u.Const.type = PrimSint;
  node->u.Const.value.i  = value;
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeConstSintTextCoord(const char *text, int value, Coord coord)
{
  Node *create = MakeConstSint(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeConstPtr(unsigned int value)
{ Node *node = NewNode(Const);
  node->u.Const.type = PtrVoid;
  node->u.Const.value.u  = value;
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeConstPtrTextCoord(const char *text, unsigned int value, Coord coord)
{
  Node *create = MakeConstPtr(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeConstUint(unsigned int value)
{ Node *node = NewNode(Const);
  node->u.Const.type = PrimUint;
  node->u.Const.value.u  = value;
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeConstUintTextCoord(const char *text, unsigned int value, Coord coord)
{
  Node *create = MakeConstUint(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeConstSlong(long value)
{ Node *node = NewNode(Const);
  node->u.Const.type = PrimSlong;
  node->u.Const.value.l  = value;
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeConstSlongTextCoord(const char *text, long value, Coord coord)
{
  Node *create = MakeConstSlong(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeConstUlong(unsigned long value)
{ Node *node = NewNode(Const);
  node->u.Const.type = PrimUlong;
  node->u.Const.value.ul  = value;
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeConstUlongTextCoord(const char *text, unsigned long value, Coord coord)
{
  Node *create = MakeConstUlong(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeConstFloat(float value)
{ Node *node = NewNode(Const);
  node->u.Const.type = PrimFloat;
  node->u.Const.value.f  = value;
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeConstFloatTextCoord(const char *text, float value, Coord coord)
{
  Node *create = MakeConstFloat(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeConstDouble(double value)
{ Node *node = NewNode(Const);
  node->u.Const.type = PrimDouble;
  node->u.Const.value.d  = value;
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeConstDoubleTextCoord(const char *text, double value, Coord coord)
{
  Node *create = MakeConstDouble(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}



GLOBAL inline Node *MakeString(const char *value)
{ Node *node = NewNode(Const),
       *adcl = MakeAdcl(EMPTY_TQ, PrimChar, MakeConstSint((signed int)(strlen(value) + 1)));

  node->u.Const.type = adcl;
  node->u.Const.value.s  = value; /* quotes stripped, escape sequences converted */
  node->u.Const.text = NULL;
  return node;
}

GLOBAL inline Node *MakeStringTextCoord(const char *text, const char *value, Coord coord)
{
  Node *create = MakeString(value);
  create->u.Const.text = text;
  create->coord = coord;
  return(create);
}


GLOBAL inline Node *MakeId(const char* text)
{
    Node *create = NewNode(Id);

    create->u.id.text = text;
    create->u.id.decl = NULL;
    return(create);
}

GLOBAL inline Node *MakeIdCoord(const char* text, Coord coord)
{
  Node *create = MakeId(text);
  create->coord = coord;
  return(create);
}



GLOBAL inline Node *MakeUnary(OpType op, Node *expr)
{ Node *create = NewNode(Unary);

  if (op == '*')
    op = INDIR;
  else if (op == '&')
    op = ADDRESS;
  else if (op == '-')
    op = UMINUS;
  else if (op == '+')
    op = UPLUS;
      
  create->u.unary.op   = op;
  create->u.unary.expr = expr;

  create->u.unary.type = NULL;
  create->u.unary.value = NULL;
  
  return create;
}

GLOBAL inline Node *MakeUnaryCoord(OpType op, Node *expr, Coord coord)
{
  Node *create = MakeUnary(op, expr);
  create->coord = coord;
  return(create);
}


GLOBAL inline Node *MakeBinop(OpType op, Node *left, Node *right)
{ Node *create = NewNode(Binop);

  create->u.binop.op    = op;
  create->u.binop.left  = left;
  create->u.binop.right = right;
  create->u.binop.type = NULL;
  create->u.binop.value = NULL;
  return(create);
}

GLOBAL inline Node *MakeBinopCoord(OpType op, Node *left, Node *right, Coord coord)
{
  Node *create = MakeBinop(op, left, right);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeCast(Node *type, Node *expr)
{ Node *create = NewNode(Cast);

  create->u.cast.type = type;
  create->u.cast.expr = expr;
  create->u.cast.value = NULL;
  return(create);
}

GLOBAL inline Node *MakeCastCoord(Node *type, Node *expr, Coord coord)
{
  Node *create = MakeCast(type, expr);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeComma(List* exprs)
{
    Node *create = NewNode(Comma);
    create->u.comma.exprs  = exprs;
    return(create);
}

GLOBAL inline Node *MakeCommaCoord(List* exprs, Coord coord)
{
  Node *create = MakeComma(exprs);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeTernary(Node *cond, Node *true, Node *false)
{
    Node *create = NewNode(Ternary);
    create->u.ternary.cond  = cond;
    create->u.ternary.true  = true;
    create->u.ternary.false = false;
    create->u.ternary.colon_coord = UnknownCoord;
    create->u.ternary.type  = NULL;
    create->u.ternary.value = NULL;
    return(create);
}

GLOBAL inline Node *MakeTernaryCoord(Node *cond, Node *true, Node *false, Coord qmark_coord, Coord colon_coord)
{
  Node *create = MakeTernary(cond, true, false);
  create->coord = qmark_coord;
  create->u.ternary.colon_coord = colon_coord;
  return(create);
}

GLOBAL inline Node *MakeArray(Node *name, List* dims)
{
    Node *create = NewNode(Array);
    create->u.array.type = NULL;
    create->u.array.name = name;
    create->u.array.dims = dims;
    return(create);
}

GLOBAL inline Node *MakeArrayCoord(Node *name, List* dims, Coord coord)
{
  Node *create = MakeArray(name, dims);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeCall(Node *name, List* args)
{
    Node *create = NewNode(Call);
    create->u.call.name = name;
    create->u.call.args = args;
    create->coord = name->coord;
    return(create);
}

GLOBAL inline Node *MakeCallCoord(Node *name, List* args, Coord coord)
{
  Node *create = MakeCall(name, args);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeInitializer(List* exprs)
{
    Node *create = NewNode(Initializer);
    create->u.comma.exprs  = exprs;
    return(create);
}

GLOBAL inline Node *MakeInitializerCoord(List* exprs, Coord coord)
{
  Node *create = MakeInitializer(exprs);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeImplicitCast(Node *type, Node *expr)
{
  Node *create = NewNode(ImplicitCast);

  create->u.implicitcast.type = type;
  create->u.implicitcast.expr = expr;
  create->u.implicitcast.value = NULL;
  return(create);
}

GLOBAL inline Node *MakeImplicitCastCoord(Node *type, Node *expr, Coord coord)
{
  Node *create = MakeImplicitCast(type, expr);
  create->coord = coord;
  return(create);
}

/*************************************************************************/
/*                                                                       */
/*                          Statement nodes                              */
/*                                                                       */
/*************************************************************************/

GLOBAL inline Node *MakeLabel(const char* name, Node *stmt)
{
    Node *create = NewNode(Label);
    create->u.label.name       = name;
    create->u.label.stmt       = stmt;
    create->u.label.references = NULL;
    return(create);
}

GLOBAL inline Node *MakeLabelCoord(const char* name, Node *stmt, Coord coord)
{
  Node *create = MakeLabel(name, stmt);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeSwitch(Node *expr, Node *stmt, List* cases)
{
    Node *create = NewNode(Switch);
    create->u.Switch.expr = expr;
    create->u.Switch.stmt = stmt;
    create->u.Switch.cases = cases;
    create->u.Switch.has_default = FALSE;
    while (cases) {
      Node *n = FirstItem(cases);
      assert(n);
      if (n->typ == Default)
	create->u.Switch.has_default = TRUE;
      cases = Rest(cases);
    }
    return(create);
}

GLOBAL inline Node *MakeSwitchCoord(Node *expr, Node *stmt, List* cases, Coord coord)
{
  Node *create = MakeSwitch(expr, stmt, cases);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeCase(Node *expr, Node *stmt, Node *container)
{
    Node *create = NewNode(Case);
    create->u.Case.expr = expr;
    create->u.Case.stmt = stmt;
    create->u.Case.container = container;
    return(create);
}

GLOBAL inline Node *MakeCaseCoord(Node *expr, Node *stmt, Node *container, Coord coord)
{
  Node *create = MakeCase(expr, stmt, container);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeDefault(Node *stmt, Node *container)
{
    Node *create = NewNode(Default);
    create->u.Default.stmt = stmt;
    create->u.Default.container = container;
    return(create);
}

GLOBAL inline Node *MakeDefaultCoord(Node *stmt, Node *container, Coord coord)
{
  Node *create = MakeDefault(stmt, container);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeIf(Node *expr, Node *stmt)
{
    Node *create = NewNode(If);
    create->u.If.expr = expr;
    create->u.If.stmt = stmt;
    return(create);
}

GLOBAL inline Node *MakeIfCoord(Node *expr, Node *stmt, Coord coord)
{
  Node *create = MakeIf(expr, stmt);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeIfElse(Node *expr, Node *true, Node *false)
{
    Node *create = NewNode(IfElse);
    create->u.IfElse.expr = expr;
    create->u.IfElse.true = true;
    create->u.IfElse.false = false;
    create->u.IfElse.else_coord = UnknownCoord;
    return(create);
}

GLOBAL inline Node *MakeIfElseCoord(Node *expr, Node *true, Node *false, Coord if_coord, Coord else_coord)
{
  Node *create = MakeIfElse(expr, true, false);
  create->coord = if_coord;
  create->u.IfElse.else_coord = else_coord;
  return(create);
}

GLOBAL inline Node *MakeWhile(Node *expr, Node *stmt)
{
    Node *create = NewNode(While);
    create->u.While.expr = expr;
    create->u.While.stmt = stmt;
    return(create);
}

GLOBAL inline Node *MakeWhileCoord(Node *expr, Node *stmt, Coord coord)
{
  Node *create = MakeWhile(expr, stmt);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeDo(Node *stmt, Node *expr)
{
    Node *create = NewNode(Do);
    create->u.Do.stmt = stmt;
    create->u.Do.expr = expr;
    create->u.Do.while_coord = UnknownCoord;
    return(create);
}

GLOBAL inline Node *MakeDoCoord(Node *stmt, Node *expr, Coord do_coord, Coord while_coord)
{
  Node *create = MakeDo(stmt, expr);
  create->coord = do_coord;
  create->u.Do.while_coord = while_coord;
  return(create);
}

GLOBAL inline Node *MakeFor(Node *init, Node *cond, Node *next, Node *stmt)
{
    Node *create = NewNode(For);
    create->u.For.init = init;
    create->u.For.cond = cond;
    create->u.For.next = next;
    create->u.For.stmt = stmt;
    return(create);
}

GLOBAL inline Node *MakeForCoord(Node *init, Node *cond, Node *next, Node *stmt, Coord coord)
{
  Node *create = MakeFor(init, cond, next, stmt);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeGoto(Node* label)
{
    Node *create = NewNode(Goto);
    create->u.Goto.label = label;
    return(create);
}

GLOBAL inline Node *MakeGotoCoord(Node* label, Coord coord)
{
  Node *create = MakeGoto(label);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeContinue(Node *container)
{
    Node *create = NewNode(Continue);
    create->u.Continue.container = container;
    return(create);
}

GLOBAL inline Node *MakeContinueCoord(Node *container, Coord coord)
{
  Node *create = MakeContinue(container);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeBreak(Node *container)
{
    Node *create = NewNode(Break);

    create->u.Break.container = container;
    return(create);
}

GLOBAL inline Node *MakeBreakCoord(Node *container, Coord coord)
{
  Node *create = MakeBreak(container);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeReturn(Node *expr)
{
    Node *create = NewNode(Return);

    create->u.Return.expr = expr;
    create->u.Return.proc = NULL;
    create->u.Return.needs_sync = FALSE;
    return(create);
}

GLOBAL inline Node *MakeReturnCoord(Node *expr, Coord coord)
{
  Node *create = MakeReturn(expr);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeBlock(Node *type, List* decl, List* stmts)
{
  Node *create = NewNode(Block);
  create->u.Block.type  = type;
  create->u.Block.decl  = decl;
  create->u.Block.stmts = stmts;
  create->u.Block.right_coord = UnknownCoord;
  return(create);
}

GLOBAL inline Node *MakeBlockCoord(Node *type, List* decl, List* stmts, Coord left_coord, Coord right_coord)
{
  Node *create = MakeBlock(type, decl, stmts);
  create->coord = left_coord;
  create->u.Block.right_coord = right_coord;
  return(create);
}

/*************************************************************************/
/*                                                                       */
/*                            Type nodes                                 */
/*                                                                       */
/*************************************************************************/

GLOBAL inline Node *MakePrim(TypeQual tq, BasicType basic)
{
    Node *create = NewNode(Prim);

    create->u.prim.tq = tq;
    create->u.prim.basic = basic;
    return(create);
}

GLOBAL inline Node *MakePrimCoord(TypeQual tq, BasicType basic, Coord coord)
{
  Node *create = MakePrim(tq, basic);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeTdef(TypeQual tq, const char *name)
{
  Node *create = NewNode(Tdef);
  create->u.tdef.name = name;
  create->u.tdef.tq = tq;
  create->u.tdef.type = NULL;
  return(create);
}

GLOBAL inline Node *MakeTdefCoord(TypeQual tq, const char *name, Coord coord)
{
  Node *create = MakeTdef(tq, name);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakePtr(TypeQual tq, Node *type)
{
    Node *create = NewNode(Ptr);
    create->u.ptr.tq = tq;
    create->u.ptr.type = type;
    return(create);
}

GLOBAL inline Node *MakePtrCoord(TypeQual tq, Node *type, Coord coord)
{
  Node *create = MakePtr(tq, type);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeAdcl(TypeQual tq, Node *type, Node *dim)
{
    Node *create = NewNode(Adcl);
    create->u.adcl.tq = tq;
    create->u.adcl.type = type;
#if 0
/* fix: we need to constant-fold dim during the parse phase in order to
   compare the types of multiply-declared arrays, but this probably isn't
    the best place to do it. -- rcm */
    create->u.adcl.dim  = SemCheckNode(dim);
#endif
    create->u.adcl.dim = dim;
    create->u.adcl.size = 0;
    return(create);
}

GLOBAL inline Node *MakeAdclCoord(TypeQual tq, Node *type, Node *dim, Coord coord)
{
  Node *create = MakeAdcl(tq, type, dim);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeFdcl(TypeQual tq, List* args, Node *returns)
{
    Node *create = NewNode(Fdcl);
    create->u.fdcl.tq = tq;
    create->u.fdcl.args = args;
    create->u.fdcl.returns = returns;
    return(create);
}

GLOBAL inline Node *MakeFdclCoord(TypeQual tq, List* args, Node *returns, Coord coord)
{
  Node *create = MakeFdcl(tq, args, returns);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeSdcl(TypeQual tq, SUEtype* type)
{
    Node *create = NewNode(Sdcl);
    create->u.sdcl.tq   = tq;
    create->u.sdcl.type = type;
    return(create);
}

GLOBAL inline Node *MakeSdclCoord(TypeQual tq, SUEtype* type, Coord coord)
{
  Node *create = MakeSdcl(tq, type);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeUdcl(TypeQual tq, SUEtype* type)
{
    Node *create = NewNode(Udcl);
    create->u.udcl.tq   = tq;
    create->u.udcl.type = type;
    return(create);
}

GLOBAL inline Node *MakeUdclCoord(TypeQual tq, SUEtype* type, Coord coord)
{
  Node *create = MakeUdcl(tq, type);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeEdcl(TypeQual tq, SUEtype* type)
{
    Node *create = NewNode(Edcl);
    create->u.edcl.tq = tq;
    create->u.edcl.type = type;
    return(create);
}

GLOBAL inline Node *MakeEdclCoord(TypeQual tq, SUEtype* type, Coord coord)
{
  Node *create = MakeEdcl(tq, type);
  create->coord = coord;
  return(create);
}


/*************************************************************************/
/*                                                                       */
/*                      Other nodes (decls et al.)                       */
/*                                                                       */
/*************************************************************************/

GLOBAL inline Node *MakeDecl(const char *name, TypeQual tq, Node *type, Node *init, Node *bitsize)
{
  Node *create = NewNode(Decl);
  create->u.decl.name = name;
  create->u.decl.tq = tq;
  create->u.decl.type = type;
  create->u.decl.init = init;
  create->u.decl.bitsize = bitsize;
  create->u.decl.references = 0;
  create->u.decl.attribs = NULL;
/* Cilk additions */
  create->u.decl.kind = 0;
  create->u.decl.scope = NULL;
  return(create);
}

GLOBAL inline Node *MakeDeclCoord(const char *name, TypeQual tq, Node *type, Node *init, Node *bitsize, Coord coord)
{
  Node *create = MakeDecl(name, tq, type, init, bitsize);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeAttrib(const char *name, Node *arg)
{
  Node *create = NewNode(Attrib);
  create->u.attrib.name = name;
  create->u.attrib.arg = arg;
  return create;
}

GLOBAL inline Node *MakeAttribCoord(const char *name, Node *arg, Coord coord)
{
  Node *create = MakeAttrib(name, arg);
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeProc(Node *decl, Node *body)
{
    Node *create = NewNode(Proc);
    create->u.proc.decl = decl;
    create->u.proc.body = body;
    create->u.proc.needs_sync = FALSE;
    create->u.proc.needs_return = FALSE;
    return(create);
}

GLOBAL inline Node *MakeProcCoord(Node *decl, Node *body, Coord coord)
{
    Node *create = MakeProc(decl, body);
    create->coord = coord;
    return(create);
}

GLOBAL inline Node *MakeText(const char *text, Bool start_new_line)
{
  Node *create = NewNode(Text);
  create->u.text.text = text;
  create->u.text.start_new_line = start_new_line;
  return create;
}

GLOBAL inline Node *MakeTextCoord(const char *text, Bool start_new_line, Coord coord)
{
  Node *create = MakeText(text, start_new_line);
  create->coord = coord;
  return create;
}


/*************************************************************************/
/*                                                                       */
/*                            GCC asm Extensions                         */
/*                                                                       */
/*************************************************************************/

GLOBAL inline Node *MakeAsm(TypeQual tq, Node *template, List *output, List *input, List *clobbered)
{
    Node *create = NewNode(Asm);
    create->u.Asm.tq = tq;
    create->u.Asm.template = template;
    create->u.Asm.output = output;
    create->u.Asm.input = input;
    create->u.Asm.clobbered = clobbered;
    return(create);
}

GLOBAL inline Node *MakeAsmCoord(TypeQual tq, Node *template, List *output, List *input, List *clobbered, Coord coord)
{
  Node *create = MakeAsm(tq, template, output, input, clobbered);
  create->coord = coord;
  return(create);
}

GLOBAL inline Node *MakeAsmArg(Node *constraint, Node *expr)
{
  Node *create = NewNode(AsmArg);
  create->u.AsmArg.constraint = constraint;
  create->u.AsmArg.expr = expr;
  return(create);
}
GLOBAL inline Node *MakeAsmArgCoord(Node *constraint, Node *expr, Coord coord)
{
  Node *create = MakeAsmArg(constraint, expr);
  create->coord = coord;
  return(create);
}

/*************************************************************************/
/*                                                                       */
/*                            Extensions                                 */
/*                                                                       */
/*************************************************************************/

GLOBAL inline Node *MakeSpawn(Node *receiver, OpType assign_op, Node *name, List *args, Bool in_inlet)
{
  Node *create = NewNode(Spawn);
  create->u.spawn.receiver = receiver;
  create->u.spawn.assign_op = assign_op;
  create->u.spawn.name = name;
  create->u.spawn.args = args;
  create->u.spawn.in_inlet = in_inlet;

  return create;
}

GLOBAL inline Node *MakeSpawnCoord(Node *receiver, OpType assign_op, Node *name, List *args, Bool in_inlet, Coord coord)
{
  Node *create = MakeSpawn(receiver, assign_op, name, args, in_inlet);
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeSync()
{
  Node *create = NewNode(Sync);
  return create;
}

GLOBAL inline Node *MakeSyncCoord(Coord coord)
{
  Node *create = MakeSync();
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeInletCall(Node *receiver, Node *name, List *args, Bool implicit_inlet, Node *binop)
{
  Node *create = NewNode(InletCall);
  create->u.inletcall.receiver = receiver;
  create->u.inletcall.name = name;
  create->u.inletcall.args = args;
  create->u.inletcall.implicit_inlet = implicit_inlet;
  create->u.inletcall.binop = binop;
  return create;
}

GLOBAL inline Node *MakeInletCallCoord(Node *receiver, Node *name, List *args, Bool implicit_inlet, Node *binop, Coord coord)
{
  Node *create = MakeInletCall(receiver, name, args, implicit_inlet, binop);
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeResume(Node *expr)
{
  Node *create = NewNode(Resume);
  create->u.resume.expr = expr;
  return create;
}

GLOBAL inline Node *MakeResumeCoord(Node *expr, Coord coord)
{
  Node *create = MakeResume(expr);
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeAbort()
{
  Node *create = NewNode(Abort);
  return create;
}

GLOBAL inline Node *MakeAbortCoord(Coord coord)
{
  Node *create = MakeAbort();
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeDoubleReturn(Node *expr)
{
  Node *create = NewNode(DoubleReturn);

  /* remove the enclosing parentheses around expr if any,
     this is because crlf (double return) is transformed
     into a call node, which will add parentheses.
   */
  if (expr) expr->parenthesized = 0;
  create->u.Return.expr = expr;
  create->u.Return.proc = NULL;
  return create;
}

GLOBAL inline Node *MakeDoubleReturnCoord(Node *expr, Coord coord)
{
  Node *create = MakeDoubleReturn(expr);
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeOnAbort(Node *stmt)
{
  Node *create = NewNode(OnAbort);

  create->u.onabort.stmt = stmt;
  return create;
}

GLOBAL inline Node *MakeOnAbortCoord(Node *stmt, Coord coord)
{
  Node *create = MakeOnAbort(stmt);
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeSynched()
{
  Node *create = NewNode(Synched);
  return create;
}

GLOBAL inline Node *MakeSynchedCoord(Coord coord)
{
  Node *create = MakeSynched();
  create->coord = coord;
  return create;
}

GLOBAL inline Node *MakeSpawnCounter()
{
  Node *create = NewNode(SpawnCounter);
  return create;
}

GLOBAL inline Node *MakeSpawnCounterCoord(Coord coord)
{
  Node *create = MakeSpawnCounter();
  create->coord = coord;
  return create;
}




/*****************************************************************
 
                Converting nodes between types

*****************************************************************/

GLOBAL Node *ConvertIdToTdef(Node *id, TypeQual tq, Node *type)
{
    assert(id->typ == Id);
    id->typ = Tdef;
    id->u.tdef.name = id->u.id.text;
    id->u.tdef.tq = tq;
    id->u.tdef.type = type;
    return(id);
}

GLOBAL Node *ConvertIdToDecl(Node *id, TypeQual tq, Node *type, Node *init, Node *bitsize)
{
    assert(id->typ == Id);
    id->typ = Decl;
    id->u.decl.name = id->u.id.text;
    id->u.decl.tq = tq;
    id->u.decl.type = type;
    id->u.decl.init = init;
    id->u.decl.bitsize = bitsize;
    id->u.decl.references = 0;
    return(id);
}

GLOBAL Node *ConvertIdToAttrib(Node *id, Node *arg)
{
    assert(id->typ == Id);
    id->typ = Attrib;
    id->u.attrib.name = id->u.id.text;
    id->u.attrib.arg = arg;
    return(id);
}


#if 0
/* dead code -- rcm */
GLOBAL Node *AdclFdclToPtr(Node *node)
{
  if (node->typ == Decl)
    if (node->u.decl.type->typ == Adcl)
      node->u.decl.type->typ = Ptr;
    else if (node->u.decl.type->typ == Fdcl)
      node->u.decl.type = MakePtrCoord(EMPTY_TQ, node->u.decl.type, node->coord);

  return node;
}
#endif




/*****************************************************************
 
                      Node-kind predicates

*****************************************************************/

GLOBAL inline Bool IsExpr(Node *node)
{ return KindsOfNode(node) & KIND_EXPR; }

GLOBAL inline Bool IsStmt(Node *node)
{ return KindsOfNode(node) & KIND_STMT; }

GLOBAL inline Bool IsType(Node *node)
{ return KindsOfNode(node) & KIND_TYPE; }

GLOBAL inline Bool IsDecl(Node *node)
{ return KindsOfNode(node) & KIND_DECL; }


/*************************************************************************/
/*                                                                       */
/*                          Expression nodes                             */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Kinds KindsOfConst (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfId (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfBinop (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfUnary (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfCast (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfComma (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfTernary (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfArray (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfCall (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfInitializer (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfImplicitCast (void)
{ return KIND_EXPR | KIND_STMT; }

/*************************************************************************/
/*                                                                       */
/*                          Statement nodes                              */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Kinds KindsOfLabel (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfSwitch (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfCase (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfDefault (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfIf (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfIfElse (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfWhile (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfDo (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfFor (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfGoto (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfContinue (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfBreak (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfReturn (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfBlock (void)
{ return ANSIOnly ? KIND_STMT : KIND_STMT | KIND_EXPR; }


/*************************************************************************/
/*                                                                       */
/*                             Type nodes                                */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Kinds KindsOfPrim (void)
{ return KIND_TYPE; }

PRIVATE inline Kinds KindsOfTdef (void)
{ return KIND_TYPE; }

PRIVATE inline Kinds KindsOfPtr (void)
{ return KIND_TYPE; }

PRIVATE inline Kinds KindsOfAdcl (void)
{ return KIND_TYPE; }

PRIVATE inline Kinds KindsOfFdcl (void)
{ return KIND_TYPE; }

PRIVATE inline Kinds KindsOfSdcl (void)
{ return KIND_TYPE | KIND_DECL; }

PRIVATE inline Kinds KindsOfUdcl (void)
{ return KIND_TYPE | KIND_DECL; }

PRIVATE inline Kinds KindsOfEdcl (void)
{ return KIND_TYPE | KIND_DECL; }

/*************************************************************************/
/*                                                                       */
/*                      Other nodes (decls et al.)                       */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Kinds KindsOfDecl (void)
{ return KIND_DECL; }

PRIVATE inline Kinds KindsOfAttrib (void)
{ return 0; }

PRIVATE inline Kinds KindsOfProc (void)
{ return KIND_DECL; }

PRIVATE inline Kinds KindsOfText (void)
{ return KIND_DECL | KIND_STMT; }


/*************************************************************************/
/*                                                                       */
/*                            GCC asm Extensions                         */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Kinds KindsOfAsm (void) 
{ return KIND_STMT; }
PRIVATE inline Kinds KindsOfAsmArg (void) 
{ return KIND_EXPR; }


/*************************************************************************/
/*                                                                       */
/*                            Extensions                                 */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Kinds KindsOfSync (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfSpawn (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfInletCall (void) 
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfResume (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfAbort (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfDoubleReturn (void)
{ return KIND_STMT; }

PRIVATE inline Kinds KindsOfOnAbort (void)
{ return KIND_DECL; }

PRIVATE inline Kinds KindsOfSynched (void)
{ return KIND_EXPR | KIND_STMT; }

PRIVATE inline Kinds KindsOfSpawnCounter (void)
{ return KIND_EXPR | KIND_STMT; }



/*************************************************************************/
/*                                                                       */
/*                           KindsOfNode                                 */
/*                                                                       */
/*************************************************************************/

GLOBAL inline Kinds KindsOfNode(Node *node)
{
#define CODE(name, node, union) return KindsOf##name()
  ASTSWITCH(node, CODE);
#undef CODE
    
  return 0; /* unreachable -- eliminates warning */
}  



/*************************************************************************/
/*                                                                       */
/*                           AST Operations                              */
/*                                                                       */
/*************************************************************************/

GLOBAL Node *NodeCopy(Node *from, TreeOpDepth d)
{
  Node *new = NewNode(from->typ);
  *new = *from;

  switch (from->typ) {
    /* for nodes with sub-lists, make new copy of sub-list */
  case Comma:         new->u.comma.exprs = ListCopy(new->u.comma.exprs);  break;
  case Array:         new->u.array.dims = ListCopy(new->u.array.dims);  break;
  case Call:          new->u.call.args = ListCopy(new->u.call.args);  break;
  case Initializer:   new->u.initializer.exprs = ListCopy(new->u.initializer.exprs);  break;
  case Block:         new->u.Block.decl = ListCopy(new->u.Block.decl);  new->u.Block.stmts = ListCopy(new->u.Block.stmts);  break;
  case Fdcl:          new->u.fdcl.args = ListCopy(new->u.fdcl.args);  break;

  case Sdcl:
  case Udcl:
  case Edcl:
    /* if from elaborated a struct/union's fields, new should not,
       since it's sharing */
    if (d == NodeOnly || d == Subtree)
      NodeRemoveTq(new, T_SUE_ELABORATED);
    break;
  case Spawn:          new->u.spawn.args = ListCopy(new->u.spawn.args);  break;
  case Asm:            new->u.Asm.output = ListCopy(new->u.Asm.output);  new->u.Asm.input = ListCopy(new->u.Asm.input);  new->u.Asm.clobbered = ListCopy(new->u.Asm.clobbered);  break;
  default:
    break;
  }

  if (d == Subtree || d == SubtreeExact) {
    /* recursively copy children */
#define CHILD(n)   n = NodeCopy(n, d)
    ASTWALK(new, CHILD)
#undef CHILD
  }

  return new;
}



PRIVATE void SetCoordsNode(Node *node, Coord *c)
{
  node->coord = *c;

  /* handle special-case coordinates as well */
  switch (node->typ) {
  case Ternary: node->u.ternary.colon_coord = *c; break;
  case IfElse:  node->u.IfElse.else_coord = *c; break;
  case Do:      node->u.Do.while_coord = *c; break;
  case Block:   node->u.Block.right_coord = *c; break;
  case Sdcl:    
    if (SUE_ELABORATED(node->u.sdcl.tq)) {
      node->u.sdcl.type->coord = *c; 
      node->u.sdcl.type->right_coord = *c;
    }
    break;
  case Udcl:
    if (SUE_ELABORATED(node->u.udcl.tq)) {
      node->u.udcl.type->coord = *c; 
      node->u.udcl.type->right_coord = *c;
    }
    break;
  case Edcl:
    if (SUE_ELABORATED(node->u.edcl.tq)) {
      node->u.edcl.type->coord = *c; 
      node->u.edcl.type->right_coord = *c;
    }
    break;
  default:
    break;
  }
}


/* SetCoords sets all coordinates on a node or subtree to c. */
GLOBAL void SetCoords(Node *node, Coord c, TreeOpDepth d)
{
  if (d == NodeOnly)
    SetCoordsNode(node, &c);
  else WalkTree(node, (WalkProc)SetCoordsNode, &c, Preorder);

  if (d == Subtree) {
    #define CHILD(n)   SetCoords(n, c, d)
    ASTWALK(node, CHILD);
    #undef CHILD
  }
}


GLOBAL void WalkTree(Node *node, WalkProc proc, void *ptr, WalkOrder order)
{
  if (order == Preorder)
    proc(node, ptr);

  #define CHILD(n)   WalkTree(n, proc, ptr, order)
  ASTWALK(node, CHILD);
  #undef CHILD
  
  if (order == Postorder)
    proc(node, ptr);
}

