/*
 * Matrix Multiply: A dag-consistent shared memory example.
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdlib.h>
#include <stdio.h>
#include <alloca.h>
#include <sys/time.h>

/* Define a block as a BLOCK_EDGE x BLOCK_EDGE array of doubles.
 */
#define BLOCK_EDGE 16
#define BLOCK_SIZE (BLOCK_EDGE*BLOCK_EDGE)
typedef double block[BLOCK_SIZE];


/* compute R = AB, where R,A,B are BLOCK_EDGE x BLOCK_EDGE matricies */
void serialmul(block *A, block *B, block *R)
{
     int i, j;

     for (j = 0; j < 16; j += 2) {	/* 2 columns at a time */
	  double *bp = &((double*)B)[j];
	  for (i = 0; i < 16; i += 2) {		/* 2 rows at a time */
	       double *ap = &((double*)A)[i * 16];
	       double *rp = &((double*)R)[j+i*16];
	       register double s0_0 = 0.0, s0_1 = 0.0;
	       register double s1_0 = 0.0, s1_1 = 0.0;
	       s0_0 += ap[0] * bp[0];
	       s0_1 += ap[0] * bp[1];
	       s1_0 += ap[16] * bp[0];
	       s1_1 += ap[16] * bp[1];
	       s0_0 += ap[1] * bp[16];
	       s0_1 += ap[1] * bp[17];
	       s1_0 += ap[17] * bp[16];
	       s1_1 += ap[17] * bp[17];
	       s0_0 += ap[2] * bp[32];
	       s0_1 += ap[2] * bp[33];
	       s1_0 += ap[18] * bp[32];
	       s1_1 += ap[18] * bp[33];
	       s0_0 += ap[3] * bp[48];
	       s0_1 += ap[3] * bp[49];
	       s1_0 += ap[19] * bp[48];
	       s1_1 += ap[19] * bp[49];
	       s0_0 += ap[4] * bp[64];
	       s0_1 += ap[4] * bp[65];
	       s1_0 += ap[20] * bp[64];
	       s1_1 += ap[20] * bp[65];
	       s0_0 += ap[5] * bp[80];
	       s0_1 += ap[5] * bp[81];
	       s1_0 += ap[21] * bp[80];
	       s1_1 += ap[21] * bp[81];
	       s0_0 += ap[6] * bp[96];
	       s0_1 += ap[6] * bp[97];
	       s1_0 += ap[22] * bp[96];
	       s1_1 += ap[22] * bp[97];
	       s0_0 += ap[7] * bp[112];
	       s0_1 += ap[7] * bp[113];
	       s1_0 += ap[23] * bp[112];
	       s1_1 += ap[23] * bp[113];
	       s0_0 += ap[8] * bp[128];
	       s0_1 += ap[8] * bp[129];
	       s1_0 += ap[24] * bp[128];
	       s1_1 += ap[24] * bp[129];
	       s0_0 += ap[9] * bp[144];
	       s0_1 += ap[9] * bp[145];
	       s1_0 += ap[25] * bp[144];
	       s1_1 += ap[25] * bp[145];
	       s0_0 += ap[10] * bp[160];
	       s0_1 += ap[10] * bp[161];
	       s1_0 += ap[26] * bp[160];
	       s1_1 += ap[26] * bp[161];
	       s0_0 += ap[11] * bp[176];
	       s0_1 += ap[11] * bp[177];
	       s1_0 += ap[27] * bp[176];
	       s1_1 += ap[27] * bp[177];
	       s0_0 += ap[12] * bp[192];
	       s0_1 += ap[12] * bp[193];
	       s1_0 += ap[28] * bp[192];
	       s1_1 += ap[28] * bp[193];
	       s0_0 += ap[13] * bp[208];
	       s0_1 += ap[13] * bp[209];
	       s1_0 += ap[29] * bp[208];
	       s1_1 += ap[29] * bp[209];
	       s0_0 += ap[14] * bp[224];
	       s0_1 += ap[14] * bp[225];
	       s1_0 += ap[30] * bp[224];
	       s1_1 += ap[30] * bp[225];
	       s0_0 += ap[15] * bp[240];
	       s0_1 += ap[15] * bp[241];
	       s1_0 += ap[31] * bp[240];
	       s1_1 += ap[31] * bp[241];
	       rp[0] = s0_0;
	       rp[1] = s0_1;
	       rp[16] = s1_0;
	       rp[17] = s1_1;
	  }
     }
}


void serialadd(block *T, block *R)
{
  long i;
  
  for(i=0;i<BLOCK_SIZE;i+=4)
    {
      ((double*)R)[i] += ((double*)T)[i];
      ((double*)R)[i+1] += ((double*)T)[i+1];
      ((double*)R)[i+2] += ((double*)T)[i+2];
      ((double*)R)[i+3] += ((double*)T)[i+3];
    }
}

/* add matrix T into matrix R */
void matrixadd(long num, block *T, block *R)
{
  if (num==1)
    {
      /* add block T to block R */
      serialadd(T, R);
    }
  else
    {
      matrixadd(num/2, T, R);
      matrixadd(num/2, T + num/2, R + num/2);
    }
  
  return;
}

/*  Blocked matrix multiply is done as follows:
 *
 *      A      x      B      =          R
 *
 *  | C | D |     | G | H |     | CG+DI | CH+DJ |
 *  |---+---|  x  |---+---|  =  |-------+-------|
 *  | E | F |     | I | J |     | EG+FI | EH+FJ |
 *
 */
/* compute R = AB, where A,B,R are nb x nb blocks in size */
void matrixmul(long nb, block *A, block *B, block *R)
{
  if (nb==1)
    {
      serialmul(A, B, R);
    }
  else
    {
      long nb2 = nb/2;
      long nb22 = nb2*nb2;
      block *C,*D,*E,*F,*G,*H,*I,*J;
      block *CG,*CH,*EH,*EG,*DI,*DJ,*FJ,*FI;
      block *tmp = malloc(nb*nb*sizeof(block));
      
      /* names for portions of original matricies */
      C = A;
      D = A + nb22;
      E = A + 3*nb22;
      F = A + 2*nb22;
      G = B;
      H = B + nb22;
      I = B + 3*nb22;
      J = B + 2*nb22;
      
      /* places to put results */
      CG = R;
      CH = CG + nb22;
      EH = CH + nb22;
      EG = EH + nb22;
      
      DI = tmp;
      DJ = DI + nb22;
      FJ = DJ + nb22;
      FI = FJ + nb22;
      
      matrixmul(nb2, C, G, CG);
      matrixmul(nb2, C, H, CH);
      matrixmul(nb2, E, H, EH);
      matrixmul(nb2, E, G, EG);
      matrixmul(nb2, D, I, DI);
      matrixmul(nb2, D, J, DJ);
      matrixmul(nb2, F, J, FJ);
      matrixmul(nb2, F, I, FI);
      
      matrixadd(nb*nb, tmp, R);
      free(tmp);
    }
  
  return;
}

void init_matrix(block *A, long num)
{
  if (num==1)
    {
      int i;
      
      for(i=0;i<BLOCK_SIZE;i++)
	((double*)A)[i]=1.0;
    }
  else
    {
      init_matrix(A, num/2);
      init_matrix(A + num/2, num/2);
    }
  
  return;
}

void check_result(block *R, double val, long num)
{
  if (num==1)
    {
      int i;
      
      for(i=0;i<BLOCK_SIZE;i++)
	if (((double*)R)[i] != val)
	  {
	    printf("bad result %4d %8f\n", i, ((double*)R)[i]);
	    return;
	  }
    }
  else
    {
      check_result(R, val, num/2);
      check_result(R + num/2, val, num/2);
    }
  
  return;
}

int main(int argc, char *argv[])
{
  long size;
  long n;
  long nblocks;
  block *A, *B, *R;
  hrtime_t start = gethrtime();
  
  if (argc != 2)
    {
      printf("Usage: %s size\n", argv[0]);
      return 1;
    }
  
  size = atoi(argv[1]);
  
  /* a size of zero indicates that the input matrices should be read
   * from an input file.
   */
  if (!size)
    {
      FILE *f;
      f = fopen("inputsize","r");
      fscanf(f, "%ld", &n);
      fclose(f);
    }
  else
    n = size;
  
  if ((n & (n-1)) != 0 || (n % BLOCK_EDGE) != 0)
    {
      printf("%ld: matrix size must be a power of 2"
	     " and a multiple of %d\n", n, BLOCK_EDGE);
      return 1;
    }
  
  nblocks = (n*n)/BLOCK_SIZE;
  
  A = malloc(nblocks * sizeof(block));
  B = malloc(nblocks * sizeof(block));
  R = malloc(nblocks * sizeof(block));
  
  if (!size)
    {
      FILE *f;
      printf("reading input matricies...\n");
      f = fopen("inputdata","r");
      
      fread((char*)A, sizeof(block), nblocks, f);
      fread((char*)B, sizeof(block), nblocks, f);
      
      fclose(f);
      printf("done reading\n");
    }
  else
    {
      init_matrix(A, nblocks);
      init_matrix(B, nblocks);
    }
  
  /* do multiplication! */
  matrixmul(n/BLOCK_EDGE, A, B, R);
  
  if (!size)
    {
      FILE *f;
      printf("writing output matrix...\n");
      f = fopen("resultdata", "w");
      
      fwrite((char*)R, sizeof(block), nblocks, f);
      
      fclose(f);
      printf("done writing\n");
    }
  else
    {
      check_result(R, n, nblocks);
    }
  
  printf("running time: %fs\n", (gethrtime()-start)/1.0E9);

  return 0;
}
