/*
 * $Id: arch-irix.c,v 1.12 1996/09/11 18:36:43 randall Exp $
 */
/*
 * $Log: arch-irix.c,v $
 * Revision 1.12  1996/09/11 18:36:43  randall
 * Updated arch-irix files.
 *
 * Revision 1.11  1996/09/10 19:43:04  randall
 * Moved bstore files to /tmp
 *
 * Revision 1.10  1996/09/07 18:33:13  randall
 * Cleaned up debug printing a little bit.  Sent debug output to stderr.
 *
 * Revision 1.9  1996/08/27 20:36:46  randall
 * Better error message if child exit()'s with nonzero value.
 *
 * Revision 1.8  1996/08/26 22:16:27  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 * Revision 1.7  1996/06/07 17:55:36  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.6  1996/06/05 15:37:35  athena
 * Added cilk-lib.h, and updated stuff
 *
 * Revision 1.5  1996/06/04 15:09:41  athena
 * Support for sparc V8 and V9
 *
 * Revision 1.4  1996/06/03 21:20:39  athena
 * Minor cleanup
 *
 * Revision 1.3  1996/05/30 20:10:27  athena
 * Added copyright legalese
 *
 * Revision 1.2  1996/05/28 21:47:48  randall
 * Updated IRIX architecture files.
 *
 * Revision 1.1  1996/05/20 16:55:29  randall
 * Added irix architecture-specific files.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk-internal.h>

/***********************************************************\
 * Timing stuff
\***********************************************************/

#if CILK_TIMING

#warning Timing statistics are not implemented for IRIX port.

#endif

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>
#include <sys/resource.h>

/* array of children's process IDs - init and set by Cilk_begin_parallel */
static pid_t *children_pid;

/*
 * Start forking processes: one process enters this function,
 * <Cilk_active_size> processes should exit this function immediately
 * and return 0.  Each returning process should set <Self> to a unique
 * number between 0 and <Cilk_active_size> - 1.  After the returning
 * processes complete (via Cilk_end_parallel()), the original process
 * should exit this function with a non-zero return value.
 */
int Cilk_begin_parallel(void)
{
     int i, n;
     pid_t pid;
     int stat;

     /* init_timer(); */

     children_pid = (pid_t *) malloc(Cilk_active_size * sizeof(pid_t));
     CILK_CHECK(children_pid,
		("couldn't malloc() children_pid\n"));

     for (i = 0; i < Cilk_active_size; i++) {
	  if ((pid = fork()) < 0) {
	       Cilk_die("can't fork");
	  } else if (pid == 0) {	/* child */
	       /* 
	        * if (processor_bind(P_PID, P_MYID, i % 8, NULL) != 0)
	        * perror("processor_bind unsuccessful"); 
	        */
	       Self = i;
	       return 0;
	  } else {		/* parent */
	       children_pid[i] = pid;
	  }
     }

     /* parent now waits for all children to complete */
     for (n = 0; n < Cilk_active_size; n++) {
	  pid = waitpid((pid_t) - 1, &stat, 0);

	  /* find completed child's # */
	  for (i = 0; i < Cilk_active_size; i++)
	       if (pid == children_pid[i])
		    break;
	  CILK_ASSERT(i < Cilk_active_size);

	  if (WIFEXITED(stat))
	      {
		if (WEXITSTATUS(stat) == 0)
		  {
		    TRACE(TRACE_MISC, ("%d: child exited normally\n", i));
		    continue;
		  }
		else
		  CILK_DPRINTF("%d: Child exited with non-zero value %d\n", i, WEXITSTATUS(stat));
	      }

	  else if (WIFSIGNALED(stat)) {
	       const char *core;
	       if (WCOREDUMP(stat))
		    core = " (core dumped)";
	       else
		    core = "";

	       /* 
	        * Print out understandable messages for common signals,
	        * and the signal number for the uncommon ones. 
	        */
	       switch (WTERMSIG(stat)) {
		   case SIGILL:
			CILK_DPRINTF("%d: Illegal instruction%s\n", i, core);
			break;
		   case SIGABRT:
			CILK_DPRINTF("%d: Abort or Assertion failed%s\n", i, core);
			break;
		   case SIGFPE:
			CILK_DPRINTF("%d: Floating exception%s\n", i, core);
			break;
		   case SIGKILL:
			CILK_DPRINTF("%d: Terminated%s\n", i, core);
			break;
		   case SIGBUS:
			CILK_DPRINTF("%d: Bus error%s\n", i, core);
			break;
		   case SIGSEGV:
			CILK_DPRINTF("%d: Segmentation fault%s\n", i, core);
			break;
		   default:
			CILK_DPRINTF("%d: signal %d (see /usr/include/sys/signal.h)%s\n", i, WTERMSIG(stat), core);
			break;
	       }
	  }
	  /* 
	   * something went wrong with child i: abort the rest of the 
	   * children.  Sending a SIGKILL keeps them from dumping core.
	   */
	  for (i = 0; i < Cilk_active_size; i++)
	       kill(children_pid[i], SIGKILL);
     }

     return 1;
}

/* Called when each child is done computing. */
void Cilk_end_parallel(void)
{
     exit(0);
}

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

#include <stdarg.h>
#include <stdio.h>

DECLARE_SHARED_VARIABLE(static LockT dprintf_lock) = 0;

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_sema_wait(&dprintf_lock);
     if (Self >= 0)
	  fprintf(stderr, "%d: ", Self);
     else
	  fprintf(stderr, "global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_sema_signal(&dprintf_lock);
}

DECLARE_SHARED_VARIABLE(static LockT die_lock) = 0;

void Cilk_die(const char *fmt,...)
{
     va_list l;

     /* ensure the first fatal error produces death of the program */
     Cilk_sema_wait(&die_lock);
     Cilk_sema_wait(&dprintf_lock);
     if (Self >= 0)
	  fprintf(stderr, "%d: fatal error: ", Self);
     else
	  fprintf(stderr, "global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_sema_signal(&dprintf_lock);
     abort();			/* dump core */
}

/***********************************************************\
 * Fault handling
\***********************************************************/

#include <signal.h>
#include <sys/ucontext.h>

DECLARE_PRIVATE_VARIABLE(static struct sigaction oldact);
DECLARE_PRIVATE_VARIABLE(static struct sigaction newact);

static void signal_handler(int sig, siginfo_t * info, ucontext_t * context)
{
     TRACE(TRACE_SIGNAL, ("got signal %d (%d)\n", sig, info->si_code));

     if (sig != SIGSEGV || info->si_code != SEGV_ACCERR ||
	 Cilk_signal_handler(info->si_addr)) {
	  /* 
	   * uninstall handler and return - user code will get real signal 
	   */
	  sigaction(SIGSEGV, &oldact, NULL);
	  return;
     }
}

void Cilk_setup_fault_handler(void)
{
     sigaction(SIGSEGV, NULL, &oldact);		/* get signal stuff */
     newact = oldact;		/* copy stuff */
     newact.sa_handler = (void (*)()) signal_handler;
     newact.sa_flags |= SA_SIGINFO;	/* modify stuff     */
     sigaction(SIGSEGV, &newact, NULL);		/* set signal stuff */
}

/***********************************************************\
 * Memory mapping
\***********************************************************/

#include <sys/mman.h>
#include <fcntl.h>

/* file descriptor of backing store file */
DECLARE_PRIVATE_VARIABLE(static int bs_fd) = -1;
DECLARE_PRIVATE_VARIABLE(static const char *bstore_name) = NULL;

/*
 * Note: all arguments of the functions below are guaranteed to be a
 * multiple of the system page size.
 */

/*
 * Allocate and set up backing store area of <len> bytes.
 */
void Cilk_setup_bstore(size_t len)
{
     /* 
      * looks like irix doesn't need a file to be full of data
      * in order to mmap() it. We will see...
      */
     /* TODO: for now put bstore in home directory */
     if (!bstore_name)
	  bstore_name = "/tmp/_Cilk_bstore";

     /* create the file */
     bs_fd = open(bstore_name, O_RDWR | O_CREAT, 0600);

     CILK_CHECK(bs_fd >= 0,
		("Open error\n"));

     /* delete temporary bstore file */
     unlink(bstore_name);
}

/*
 * Allocate a chunk of virtual memory of size <len>, and initialize it
 * to fault when referenced.  Returns pointer to the start of memory.
 */
void *Cilk_map_alloc(size_t len)
{
     void *area;

     CILK_ASSERT(bs_fd >= 0);
     area = mmap(0, len, PROT_NONE, MAP_SHARED, bs_fd, 0);
     if (area == (void *) -1)
	  Cilk_die("mmap error in Cilk_map_alloc\n");

     return area;
}

/* 
 * Make section of memory [<addr>, <addr>+<len>-1] give a page fault
 * when referenced.
 */
void Cilk_map_fault(void *addr, size_t len)
{
     CILK_ASSERT(PAGE_ALIGNED_P(addr));
     CILK_ASSERT(PAGE_ALIGNED_P(len));
#ifdef USE_MPROTECT
     /* 
      * I suspect mprotect() is broken on Solaris.
      * I think I found a case in which a page mapped PROT_NONE
      * is still accessible. Maybe the hardware is broken. -- athena
      */
     CILK_CHECK(mprotect(addr, len, PROT_NONE) == 0,
		("couldn't mprotect memory! %x %d\n", (int) addr, len));
#else
     /* 
      * we use munmap() for now. 
      */
     CILK_CHECK(munmap(addr, len) == 0,
		("couldn't munmap memory! %x %d\n", (int) addr, len));
#endif
}

/*
 * Map section of memory [<addr>, <addr>+<len>-1] to section of
 * backing store [<offset>, <offset>+<len>-1].
 */
void Cilk_map_bstore_page(void *addr, int bs_page)
{
     CILK_ASSERT(PAGE_ALIGNED_P(addr));
     if (addr != mmap(addr, Cilk_pagesize, PROT_READ | PROT_WRITE,
		      MAP_SHARED | MAP_FIXED | MAP_AUTOGROW,
		      bs_fd, bs_page * Cilk_pagesize))
	  Cilk_die("couldn't map to file.\n");
}

/***********************************************************\
 * Shared Global Variables Mapping
\***********************************************************/

/* 
 * map the global variables segment as shared
 */
void Cilk_map_global_variables(void)
{
     extern _end_Cilk_shared;	/* linker_defined symbols */
     extern _begin_Cilk_shared;
     char name[1024];
     int fd;
     void *startmem, *endmem;

     if (bstore_name) {
	  CILK_ASSERT(strlen(bstore_name) < 1000);
	  sprintf(name, "%s.variables", bstore_name);
     } else {
	  sprintf(name, "/tmp/_Cilk_gvars");
     }

     startmem = &_begin_Cilk_shared;
     endmem = &_end_Cilk_shared;

     TRACE(TRACE_SETUP,
	("startmem = %x,  endmem = %x\n", (int) startmem, (int) endmem));

     CILK_ASSERT(((int) startmem) % Cilk_pagesize == 0);
     CILK_ASSERT(((int) endmem) % Cilk_pagesize == 0);

     /* write all shared data to the file */
     fd = open(name, O_RDWR | O_CREAT, 0600);
     CILK_CHECK(fd >= 0, ("Open error\n"));
     unlink(name);
     CILK_CHECK(write(fd, startmem, endmem - startmem) == endmem - startmem,
		("error in write\n"));

     /* map shared data to be shared */
     if (startmem != mmap(startmem, endmem - startmem, PROT_READ | PROT_WRITE,
			  MAP_SHARED | MAP_FIXED, fd, 0))
	  Cilk_die("couldn't map shared part to file.\n");
}

/***********************************************************\
 * Malloc support
\***********************************************************/

/* Get a chunk of memory of size <len> that is mapped shared. */
void *Cilk_map_shared(size_t len)
{
     int fd;
     void *area;

     fd = open("/dev/zero", O_RDWR);

     CILK_CHECK(fd >= 0, ("Open error in Cilk_map_shared\n"));

     area = mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_AUTORESRV, fd, 0);
     if (area == (void *) -1)
	  Cilk_die("mmap error in Cilk_map_shared -- "
		   "couldn't allocate %ld bytes\n", len);
     close(fd);
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
     for (; i < argc; ++i) {
	  if (*argv[i] != '-')
	       break;

	  if (strcmp(argv[i] + 1, "-") == 0) {
	       ++i;
	       break;
	  } else if (strcmp(argv[i] + 1, "bfile") == 0) {
	       ++i;
	       CILK_CHECK(i < argc,
			  ("-bfile requires a file name\n"));
	       bstore_name = argv[i];
	  } else
	       break;
     }
     return i;
}
