
/*
 * $Id: arch-irix.h,v 1.12 1996/09/11 18:36:45 randall Exp $
 */
/*
 * $Log: arch-irix.h,v $
 * Revision 1.12  1996/09/11 18:36:45  randall
 * Updated arch-irix files.
 *
 * Revision 1.11  1996/08/26 22:16:29  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 * Revision 1.10  1996/06/07 22:15:55  athena
 * Fixed mistake Cilk -> Silk everywhere.
 *
 * Revision 1.9  1996/06/07 22:03:32  athena
 * Minor fix
 *
 * Revision 1.8  1996/06/07 21:42:40  athena
 * Added a couple of paranoid tests
 *
 * Revision 1.7  1996/06/07 17:55:37  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.6  1996/06/04 20:13:00  athena
 * Added Cilk-fence. SPARC V8 is now default platform.
 *
 * Revision 1.5  1996/06/03 22:14:54  athena
 * Fixed irix port
 *
 * Revision 1.4  1996/06/03 21:20:40  athena
 * Minor cleanup
 *
 * Revision 1.3  1996/05/30 20:10:28  athena
 * Added copyright legalese
 *
 * Revision 1.2  1996/05/28 21:47:50  randall
 * Updated IRIX architecture files.
 *
 * Revision 1.1  1996/05/20 16:55:31  randall
 * Added irix architecture-specific files.
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _ARCH_IRIX_H
#define _ARCH_IRIX_H

#if (!CILK_IRIX)
#error CILK_IRIX not defined!
#endif

/*
 *  If your hardware is sequentially consistent, you can define CILK_SYNC 
 *  to be a NOP.
 *  If your hardware has relaxed consistency, use the CILK_SYNC macro below.
 */
#if 0
#define CILK_SYNC asm volatile ("sync"::)
#else
#define CILK_SYNC
#endif

#if 1
#define CACHE_LINE_ALIGNED(decl) \
      decl  __attribute__((aligned(CACHE_LINE)))
#define DECLARE_SHARED_VARIABLE(decl) \
      decl __attribute__((section("Cilk_shared"), nocommon))
#define DECLARE_PRIVATE_VARIABLE(decl)  \
      decl __attribute__((section("Cilk_private"), nocommon))
#else
/* these are wrong, but necessary */
#define CACHE_LINE_ALIGNED(decl) decl
#define DECLARE_SHARED_VARIABLE(decl) decl
#define DECLARE_PRIVATE_VARIABLE(decl) decl
#endif

/***********************************************************\
 * Lock implementation
\***********************************************************/

/* a lock variable */
typedef volatile int LockT;

/* 0 = unlocked, 1 = locked */

/* initialize lock */
static inline void Cilk_sema_init(LockT *lock)
{
     *lock = 0;
     CILK_SYNC;
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_sema_wait(LockT *lock)
{
     int tmp;

     do {
	  do {
	       asm volatile ("ll %0, (%1)":"=r" (tmp):"r"(lock));
	  } while (tmp != 0);

	  tmp = 1;
	  asm volatile ("sc %0, (%1)":"=r" (tmp):"r"(lock), "0"(tmp));
     } while (tmp == 0);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_sema_try(LockT *lock)
{
     int tmp;

     asm volatile ("ll %0, (%1)":"=r" (tmp):"r"(lock));
     if (tmp != 0)
	  return 0;

     tmp = 1;
     asm volatile ("sc %0, (%1)":"=r" (tmp):"r"(lock), "0"(tmp));
     return tmp;
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_sema_signal(LockT *lock)
{
     CILK_SYNC;
     *lock = 0;
}

/*
 *  A lock type that is by itself on a cache line.
 */
typedef struct {
     CACHE_LINE_ALIGNED(LockT lock);
} AlignedLockT;

/***********************************************************\
 * Machine size
\***********************************************************/

#include <unistd.h>

static inline int Cilk_partition_size(void)
{
     /* 
      * This is only really right for an unloaded machine.
      */
     return sysconf(_SC_NPROC_ONLN);
}

/***********************************************************\
 * Memory barriers
\***********************************************************/

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
     CILK_SYNC;
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
     CILK_SYNC;
}

/*
 *  Ensure that all previous reads are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_LoadLoad(void)
{
     CILK_SYNC;
}

/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
     CILK_SYNC;
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/time.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef long long TimeT;

static inline TimeT Cilk_get_process_time(void)
{
     /* 
      * not implemented!  We couln not find a fast user-level high-resolution
      * timer. 
      */
     return 0;
}

static inline TimeT Cilk_get_wall_time(void)
{
     struct timeval tv;
     gettimeofday(&tv);
     return tv.tv_sec * (long long) 1000000 + tv.tv_usec;
}

static inline double Cilk_process_time_to_sec(TimeT t)
{
     return t / 1e6;
}

static inline double Cilk_wall_time_to_sec(TimeT t)
{
     return t / 1e6;
}

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

int Cilk_begin_parallel(void);
void Cilk_end_parallel(void);

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

void Cilk_dprintf(const char *fmt,...);
void Cilk_die(const char *fmt,...);

/***********************************************************\
 * Fault handling
\***********************************************************/

void Cilk_setup_fault_handler(void);

/***********************************************************\
 * Memory mapping
\***********************************************************/

/*
 * Get OS page size.
 */
static inline size_t Cilk_get_page_size(void)
{
     return sysconf(_SC_PAGESIZE);
}

/* set up backing store */
void Cilk_setup_bstore(size_t len);

/* set up frame stack memory area */
void *Cilk_map_alloc(size_t len);

/* manipulate page map */
void Cilk_map_fault(void *addr, size_t len);
void Cilk_map_bstore_page(void *addr, int bspage);

/*
 * Used by rts to help with global variable hack.  Should be >=
 * real OS page size.
 */
#define CILK_PAGESIZE_HACK 8192

/* global variable hack */
void Cilk_map_global_variables(void);

/* malloc support */
void *Cilk_map_shared(size_t len);

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

int Cilk_arch_specific_options(int i, int argc, char *argv[]);

#endif				/* _ARCH_IRIX_H */
