/*
 * $Id: arch-sunos.h,v 1.7 1996/09/16 17:22:14 randall Exp $
 */
/*
 * $Log: arch-sunos.h,v $
 * Revision 1.7  1996/09/16 17:22:14  randall
 * Added missing prototypes.
 *
 * Revision 1.6  1996/09/10 19:43:11  randall
 * Moved bstore files to /tmp
 *
 * Revision 1.5  1996/08/27 20:34:50  randall
 * Some SunOS support.  The linker format file still needs to be written.
 *
 * Revision 1.4  1996/08/26 22:16:35  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 * Revision 1.3  1996/07/08 22:09:36  athena
 * Random fixes for SunOS.
 *
 * Revision 1.2  1996/06/14 17:09:48  athena
 * Fixed prototype.
 *
 * Revision 1.1  1996/06/14 16:57:24  athena
 * Added support for SunOS (very crude at this point).
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _ARCH_SUNOS_H
#define _ARCH_SUNOS_H

#define SPARC_V8 1
#define SPARC_V9 0
#define OLD_GCC 1		/* pre 2.7.1 */

#if OLD_GCC
/* these are wrong, but nothing we can do until we get gcc 2.7.1 and ld 2.7... */
#define CACHE_LINE_ALIGNED(decl) decl
#define DECLARE_SHARED_VARIABLE(decl) decl
#define DECLARE_PRIVATE_VARIABLE(decl) decl
#else
#define CACHE_LINE_ALIGNED(decl) decl  __attribute__((aligned(CACHE_LINE)))
#define DECLARE_SHARED_VARIABLE(decl) \
      decl __attribute__((section("Cilk_shared"), nocommon))
#define DECLARE_PRIVATE_VARIABLE(decl)  \
      decl __attribute__((section("Cilk_private"), nocommon))
#endif

/* safety test */
#if SPARC_V8 && SPARC_V9
#error define either SPARC_V8 or SPARC_V9, but not both
#endif

#if (!SPARC_V8) && (!SPARC_V9)
#error define either SPARC_V8 or SPARC_V9
#endif

#if (!CILK_SUNOS)
#error CILK_SUNOS not defined !
#endif

/***********************************************************\
 * Lock implementation
\***********************************************************/

/* a lock variable */
typedef volatile int LockT;

/* helper function - a hardware swap instruction */
static inline int Cilk_xmem(register volatile int *ptr, register int xchg)
{
     asm volatile ("swap [%1],%0"
		   :"=r" (xchg)	/* registers written */
		   :"r"(ptr), "0"(xchg)		/* registers read */
     );
     return xchg;
}

/* initialize lock */
static inline void Cilk_sema_init(LockT *lock)
{
     Cilk_xmem(lock, 0);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_sema_wait(LockT *lock)
{
     /* 
      * This hackery should reduce bus traffic, if the
      * cache is snooped correctly.
      */
     while (Cilk_xmem(lock, 1) != 0) {
	  while (*lock != 0);
     }
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_sema_try(LockT *lock)
{
     if (*lock != 0)  /* don't try for lock until we know we might succeed */
          return 0;
     if (Cilk_xmem(lock, 1) == 0) {
	  return 1;
     } else
	  return 0;
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_sema_signal(LockT *lock)
{
     Cilk_xmem(lock, 0);
}

/*
 *  A lock type that is by itself on a cache line.
 */
typedef struct {
     CACHE_LINE_ALIGNED(LockT lock);
} AlignedLockT;

/***********************************************************\
 * Machine size
\***********************************************************/

#include <unistd.h>

static inline int Cilk_partition_size(void)
{
     /* 
      * just pretend we have 4 processors.
      */
     return 4;
}

/***********************************************************\
 * Various types of membars
\***********************************************************/

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
     /* don't need to do anything for TSO (Total Store Order) */
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
#if SPARC_V9
     asm volatile ("membar #StoreLoad"::);
#endif
#if SPARC_V8
     /* 
      * The only way I know to implement the membar on a V8 sparc
      * is to use swap. Maybe there is a better way... --- athena
      */
     int dummy;
     Cilk_xmem(&dummy, 0);
#endif
}

/*
 *  Ensure that all previous reads are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_LoadLoad(void)
{
     /* 
      * This is tricky. I *believe* that TSO guarantees order of
      * loads, but I found no clear reference anywhere.
      *
      * Here is the problem: what happens in the sync code
      * 
      *    if (join_count > 0) return;
      *    read(A)
      *
      * if some aggressive speculation is allowed to swap the read
      * of the join counter and A ? 
      *
      * So far NOP seems to work, but, if you have trouble, try (on V9) :
      */
     /* asm volatile ("membar #LoadLoad"::); */
}

/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
     /* 
      * this should be ok, maybe there is a faster way.
      * Notice that we are using SWAP only for its side effect:
      * atomic operations are ordered with respect to loads and
      * stores.
      */
     int dummy;
     Cilk_xmem(&dummy, 0);
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/time.h>
#include <sys/timeb.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef long TimeT;

static inline TimeT Cilk_get_process_time(void)
{
     long clock();
     return clock();
}

static inline TimeT Cilk_get_wall_time(void)
{
     int ftime(struct timeb *);
     struct timeb tp;
     ftime(&tp);
     return tp.time * 1000 + tp.millitm;
}

static inline double Cilk_process_time_to_sec(TimeT t)
{
     return (double) t / 1.0e6;
}

static inline double Cilk_wall_time_to_sec(TimeT t)
{
     return (double) t / 1.0e3;
}

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

int Cilk_begin_parallel(void);
void Cilk_end_parallel(void);

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

void Cilk_dprintf(const char *fmt,...);
void Cilk_die(const char *fmt,...);

/***********************************************************\
 * Fault handling
\***********************************************************/

void Cilk_setup_fault_handler(void);

/***********************************************************\
 * Memory mapping
\***********************************************************/

/*
 * Get OS page size.
 */
#include <machine/param.h>

static inline size_t Cilk_get_page_size(void)
{
     return PAGESIZE;
}

/* set up backing store */
void Cilk_setup_bstore(size_t len);

/* set up frame stack memory area */
void *Cilk_map_alloc(size_t len);

/* manipulate page map */
void Cilk_map_fault(void *addr, size_t len);
void Cilk_map_bstore_page(void *addr, int bs_page);

/*
 * Used by rts to help with global variable hack.  Should be >=
 * real OS page size.
 */
#define CILK_PAGESIZE_HACK 8192

/* global variable hack */
void Cilk_map_global_variables(void);

/* malloc support */
void *Cilk_map_shared(size_t len);

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

int Cilk_arch_specific_options(int i, int argc, char *argv[]);


/***********************************************************\
 * Missing prototypes
\***********************************************************/

#include <stdio.h>
#include <stdarg.h>

char *memalign(size_t alignment, size_t size);
char *valloc(size_t size);
int fprintf(FILE *stream, const char *format, ...);
int vfprintf(FILE *stream, const char *format, va_list ap);
int fflush(FILE *stream);
int fclose(FILE *stream);
void bcopy(const char *b1, char *b2, size_t length);
void bzero(char *b, size_t length);
long strtol(const char *str, char **ptr, int base);
int mprotect(caddr_t addr, size_t len, int prot);

#endif				/* _ARCH_SUNOS_H */
