/*
 * $Id: cilk-compat.h,v 1.12 1996/09/09 19:48:30 randall Exp $
 */
/*
 * $Log: cilk-compat.h,v $
 * Revision 1.12  1996/09/09 19:48:30  randall
 * Added rand/srand back into compat file.
 *
 * Revision 1.11  1996/08/27 17:34:46  athena
 * Added Makefile to cilklib directory
 *
 * Revision 1.10  1996/08/22 03:08:40  randall
 * Added rudimentary file-pointer-less files.  I don't expect them to
 * be a permanent addition, but they illuminate some of the I/O issues.
 * See transpose.cilk in the examples directory for usage.
 *
 * Revision 1.9  1996/08/20 22:14:27  randall
 * Added calloc, realloc, valloc.
 *
 * Revision 1.8  1996/08/20 20:59:39  randall
 * Reorganized cilk-lib.h and cilk-compat.h.  Added wrappers for Unix I/O
 * functions.
 *
 * Revision 1.7  1996/08/13 01:39:37  randall
 * Made common makefile for rts and examples directories.
 *
 * Revision 1.6  1996/06/14 16:57:25  athena
 * Added support for SunOS (very crude at this point).
 *
 * Revision 1.5  1996/06/07 17:55:42  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.4  1996/06/06 14:37:38  athena
 * cilk_main is now main
 *
 * Revision 1.3  1996/06/05 15:37:40  athena
 * Added cilk-lib.h, and updated stuff
 *
 * Revision 1.2  1996/06/04 16:32:32  athena
 * Adapted examples, because inlet_call is no longer there.
 *
 * Revision 1.1  1996/06/04 16:05:46  athena
 * Added cilk-compat.h
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_COMPAT_H
#define _CILK_COMPAT_H

/***********************************************************\
 * Memory allocator functions
\***********************************************************/

/* alloca() --- still an open problem */
#define alloca Cilk_alloca

/* malloc() and company */
#define malloc Cilk_malloc
#define memalign Cilk_memalign
#define free Cilk_free
#define calloc Cilk_calloc
#define realloc Cilk_realloc
#define valloc Cilk_valloc

/***********************************************************\
 * Random number generator
\***********************************************************/

#define rand Cilk_rand
#define srand Cilk_srand

/***********************************************************\
 * Unix I/O functions
\***********************************************************/

#define read Cilk_read
#define write Cilk_write
#define open Cilk_open
#define creat Cilk_creat
#define close Cilk_close
#define unlink Cilk_unlink
#define lseek Cilk_lseek
#define stat(x,y) Cilk_stat(x,y)
/* Note: stat must be a function macro because it takes a struct
 * whose tag is also "stat"!
 */

#endif				/* _CILK_COMPAT_H */
