/*
 * $Id: cilk-lib.h,v 1.12 1996/09/06 19:26:02 randall Exp $
 */
/*
 * $Log: cilk-lib.h,v $
 * Revision 1.12  1996/09/06 19:26:02  randall
 * Some small updates.  Added file I/O with file pointers.  All file I/O
 * stuff is #ifdef'd out for now.
 *
 * Revision 1.11  1996/08/29 15:51:29  athena
 * Added fine-grain work timings.  Abstracted subcomputations a bit.
 *
 * Revision 1.10  1996/08/27 23:11:46  randall
 * Made the file-pointer-less files operational.
 *
 * Revision 1.9  1996/08/27 20:34:52  randall
 * Some SunOS support.  The linker format file still needs to be written.
 *
 * Revision 1.8  1996/08/26 22:16:43  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 * Revision 1.7  1996/08/22 03:08:44  randall
 * Added rudimentary file-pointer-less files.  I don't expect them to
 * be a permanent addition, but they illuminate some of the I/O issues.
 * See transpose.cilk in the examples directory for usage.
 *
 * Revision 1.6  1996/08/20 22:14:29  randall
 * Added calloc, realloc, valloc.
 *
 * Revision 1.5  1996/08/20 21:06:35  randall
 * Alloca() hack now works.
 *
 * Revision 1.4  1996/08/20 20:59:40  randall
 * Reorganized cilk-lib.h and cilk-compat.h.  Added wrappers for Unix I/O
 * functions.
 *
 * Revision 1.3  1996/08/13 01:39:38  randall
 * Made common makefile for rts and examples directories.
 *
 * Revision 1.2  1996/06/07 17:55:44  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.1  1996/06/05 15:37:41  athena
 * Added cilk-lib.h, and updated stuff
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_LIB_H
#define _CILK_LIB_H

#ifdef _CILK_COMPAT_H
#error cilk_lib.h must be included before cilk_compat.h
/*
 * We should really just undef everything in cilk_compat.h,
 * and then re-include it at the end.  This would create
 * consistency problems between the files, however.
 */
#endif

/***********************************************************\
 * Exported variables
\***********************************************************/

/* Number of processors Cilk is running on */
extern int Cilk_active_size;

/* Processor index of local processor (0 <= Self < Cilk_active_size) */
extern int Self;

/***********************************************************\
 * Exported functions
\***********************************************************/

extern void Cilk_touch_mem(void *buf, size_t len);

/***********************************************************\
 * Memory allocator functions
\***********************************************************/

#include <stdlib.h>
#include <malloc.h>

/*
 * Define C_ functions to be original functions
 */
/* is this portable?  it seems to be... -KHR */
extern void *__builtin_alloca(size_t size);
#define C_alloca __builtin_alloca

static inline void *C_malloc(size_t size)
{
     return malloc(size);
}
static inline void *C_memalign(size_t alignment, size_t size)
{
     return memalign(alignment, size);
}
static inline void C_free(void *s)
{
     free(s);
}
static inline void *C_calloc(size_t nelem, size_t elsize)
{
     return calloc(nelem, elsize);
}
static inline void *C_realloc(void *s, size_t size)
{
     return realloc(s, size);
}
static inline void *C_valloc(size_t size)
{
/*     extern void *valloc(size_t size);*/
     return valloc(size);
}

extern void *Cilk_alloca(size_t size);
extern void *Cilk_malloc(size_t size);
extern void *Cilk_memalign(size_t alignment, size_t size);
extern void Cilk_free(void *s);
extern void *Cilk_calloc(size_t nelem, size_t elsize);
extern void *Cilk_realloc(void *s, size_t size);
extern void *Cilk_valloc(size_t size);

/***********************************************************\
 * Random number generator
\***********************************************************/

#include <stdlib.h>

/*
 * Define C_ functions to be original functions
 */
static inline int C_rand(void)
{
     return rand();
}
static inline void C_srand(unsigned int seed)
{
     srand(seed);
}

extern int Cilk_rand(void);
extern void Cilk_srand(unsigned int seed);

/***********************************************************\
 * Unix I/O functions
\***********************************************************/

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

/*
 * Define C_ functions to be original functions
 */
static inline int C_open(char *name, int flags, int perms)
{
     return open(name, flags, perms);
}
static inline int C_close(int fd)
{
     return close(fd);
}
static inline int C_read(int fd, char *buf, int n)
{
     return read(fd, buf, n);
}
static inline int C_write(int fd, char *buf, int n)
{
     return write(fd, buf, n);
}
static inline long C_lseek(int fd, long offset, int origin)
{
     return lseek(fd, offset, origin);
}
static inline int C_creat(char *name, int perms)
{
     return creat(name, perms);
}
static inline int C_unlink(char *name)
{
     return unlink(name);
}
static inline int C_stat(const char *name, struct stat *buf)
{
     return stat(name, buf);
}

#include <string.h>

#if CILK_FILEIO

extern int Cilk_open(const char *path, int flags, ... /* int mode */ );
extern int Cilk_close(int cilk_fd);
extern int Cilk_read(int cilk_fd, void *buf, size_t len);
extern int Cilk_write(int cilk_fd, void *buf, size_t len);
extern off_t Cilk_lseek(int cilk_fd, off_t offset, int whence);

#else    /* CILK_FILEIO */

/*
 * Cilk_ functions are just wrappers for now.  Eventually, we
 * will implement these functions from scratch...
 */
static inline int Cilk_open(char *name, int flags, int perms)
{
     strlen(name);		/* strlen() should map this string for us! 
				 * 
				 */
     return open(name, flags, perms);
}

static inline int Cilk_close(int fd)
{
     return close(fd);
}

static inline int Cilk_read(int fd, char *buf, int n)
{
     /* 
      * We should actually map and call read individually for each
      * page of the buffer, because on some Cilk implementations, we
      * won't be able to map an arbitrarily large portion of memory.
      * For our SMP version, however, we are OK for now.
      */
     Cilk_touch_mem(buf, n);
     return read(fd, buf, n);
}

static inline int Cilk_write(int fd, char *buf, int n)
{
     Cilk_touch_mem(buf, n);
     return write(fd, buf, n);
}

static inline long Cilk_lseek(int fd, long offset, int origin)
{
     return lseek(fd, offset, origin);
}

#endif    /* CILK_FILEIO */

static inline int Cilk_creat(char *name, int perms)
{
     strlen(name);		/* strlen() should map this string for us! 
				 * 
				 */
     return creat(name, perms);
}

static inline int Cilk_unlink(char *name)
{
     strlen(name);		/* strlen() should map this string for us! 
				 * 
				 */
     return unlink(name);
}

static inline int Cilk_stat(const char *name, struct stat *buf)
{
     strlen(name);		/* strlen() should map this string for us! 
				 * 
				 */
     Cilk_touch_mem(buf, sizeof(struct stat));
     return stat(name, buf);
}

#endif				/* _CILK_LIB_H */
