/*
 * $Id: cilk.c,v 1.179 1996/09/17 01:24:47 athena Exp $
 */

/*
 * $Log: cilk.c,v $
 * Revision 1.179  1996/09/17 01:24:47  athena
 * Added -infofile option
 *
 * Revision 1.178  1996/09/11 18:36:50  randall
 * Updated arch-irix files.
 *
 * Revision 1.177  1996/09/07 18:34:34  randall
 * File I/O bug fixes.
 *
 * Revision 1.176  1996/09/06 19:26:07  randall
 * Some small updates.  Added file I/O with file pointers.  All file I/O
 * stuff is #ifdef'd out for now.
 *
 * Revision 1.175  1996/09/05 19:21:49  randall
 * Small changes.
 *
 * Revision 1.174  1996/09/05 14:51:23  athena
 * Rationalization of timers
 *
 * Revision 1.173  1996/09/04 04:08:43  fengmd
 * add the preliminary version of cilk debugger
 *
 * Revision 1.172  1996/09/03 19:05:32  athena
 * Added many paranoid tests.  Using munmap() instead of mprotect()
 *
 * Revision 1.171  1996/09/03 14:59:03  athena
 * Added more debugging stuff
 *
 * Revision 1.170  1996/09/02 22:07:24  athena
 * Fixed another bug in abort code (missed cast)
 *
 * Revision 1.169  1996/09/02 19:37:04  athena
 * Fixed bug in inlet processing
 *
 * Revision 1.168  1996/08/30 18:50:43  athena
 * Polished steal code. I really wanted to do it.
 *
 * Revision 1.167  1996/08/30 16:18:12  athena
 * Fixed the 'int foo()' vs 'cilk int foo()' typechecking bug.
 *
 * Revision 1.166  1996/08/29 15:51:33  athena
 * Added fine-grain work timings.  Abstracted subcomputations a bit.
 *
 * Revision 1.165  1996/08/27 20:34:56  randall
 * Some SunOS support.  The linker format file still needs to be written.
 *
 * Revision 1.164  1996/08/27 19:28:23  randall
 * One pass of cleaning up.
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */
#include <cilk-internal.h>
#include <stdlib.h>
#include <stdio.h>

#define WAIT_A_BIT() { int i; for (i = 0; i < 1000; ++i) ; }

/***********************************************************\
 * Per-processor Variables
\***********************************************************/

/* first data segment variable */
DECLARE_PRIVATE_VARIABLE(char Cilk_result[1024]);	/* TODO */

/* user visible */
DECLARE_PRIVATE_VARIABLE(int Self) = -1;

/* cilk2c visible */
DECLARE_PRIVATE_VARIABLE(int Cilk_result_valid) = 0;
DECLARE_PRIVATE_VARIABLE(int Cilk_resume) = 0;
DECLARE_PRIVATE_VARIABLE(int Cilk_abort) = 0;
DECLARE_PRIVATE_VARIABLE(volatile StealStuffT * Cilk_ss) = NULL;
WHEN_CILK_CP(DECLARE_PRIVATE_VARIABLE(TimeT Cilk_cp_hack) = 0;)
WHEN_CILK_CP(DECLARE_PRIVATE_VARIABLE(TimeT Cilk_work_hack) = 0;)

/* rts only */
DECLARE_PRIVATE_VARIABLE(volatile int Cilk_allow_faults) = 0;
DECLARE_PRIVATE_VARIABLE(volatile int Cilk_last_mapped_page) = -1;
DECLARE_PRIVATE_VARIABLE(static volatile struct PageTableT *page_table) = NULL;

DECLARE_PRIVATE_VARIABLE(const char *Cilk_assertion_failed_msg) =
 "Assertion failed: %s line %d file %s\n"
     /* Sorry, couldn't resist --- athena */
 "Ex absurdo sequitur quodlibet.\n";

/***********************************************************\
 * Random Junk (shared variables from now on)
\***********************************************************/

DECLARE_SHARED_VARIABLE(static volatile int done);
	      /* flag set when cilk_main returns */

DECLARE_SHARED_VARIABLE(static volatile int retval) = 0;
	     /* value that cilk_main returned */

DECLARE_SHARED_VARIABLE(int CilkStatLevel) = 1;
DECLARE_SHARED_VARIABLE(long Cilk_trace_mode) = 0;

DECLARE_SHARED_VARIABLE(static TimeT start_time);
	   /* start of cilk program */
DECLARE_SHARED_VARIABLE(int Cilk_active_size) = 1;
	   /* number of processes */

DECLARE_SHARED_VARIABLE(void *Cilk_frame_stack);
	   /* area of memory for frame stack */

/* Cilk_proc_mem: size of frame stack */
DECLARE_SHARED_VARIABLE(size_t Cilk_proc_mem) = CILK_DEFAULT_STACKSIZE;

DECLARE_SHARED_VARIABLE(TimeT Cilk_critical_path);
DECLARE_SHARED_VARIABLE(TimeT Cilk_total_work);

DECLARE_SHARED_VARIABLE(char *Cilk_infofile_name) = "/dev/tty";
			/* file where info messages are sent */

FILE *Cilk_infofile;		/* neither local nor global */
size_t Cilk_pagesize;           /* size of an OS page */

/*
 * The statistics printer 
 */
static void print_all_statistics(void)
{
     Cilk_infofile = fopen(Cilk_infofile_name, "w");
     CILK_CHECK(Cilk_infofile,
		("can't open %s for writing\n", Cilk_infofile_name));

     if (CilkStatLevel >= 1) {
	  /* level 1 and above: print wall clock, work and CP */
	  fprintf(Cilk_infofile,
		  "Wall-clock running time on %d processor%s: %f\n",
		  Cilk_active_size, Cilk_active_size > 1 ? "s" : "",
	       Cilk_wall_time_to_sec(Cilk_get_wall_time() - start_time));
	  WHEN_CILK_TIMING(fprintf(Cilk_infofile, "Total work = %f\n",
				   Cilk_compute_work()));
#if CILK_CRITICAL_PATH
	  if (Cilk_critical_path != 0.0) {
	       fprintf(Cilk_infofile, "Total work (fine-grain) = %f\n",
		       Cilk_process_time_to_sec(Cilk_total_work));
	       fprintf(Cilk_infofile, "Critical path = %f\n",
		       Cilk_process_time_to_sec(Cilk_critical_path));
	       fprintf(Cilk_infofile, "Average parallelism = %f\n",
		       (double) Cilk_total_work / Cilk_critical_path);
	  }
#endif
     }
     if (CilkStatLevel >= 2) {
	  /* level 2 and above: print statistics on work stealing etc */
	  Cilk_print_rts_statistics();
     }
     if (CilkStatLevel >= 3) {
	  /* level 3 and above: print work done by each processor */
	  Cilk_print_time_statistics();
     }
     if (CilkStatLevel >= 4) {
	  /* level 4 and above: print detailed breakout of time */
	  Cilk_print_detailed_time_statistics();
     }
     fclose(Cilk_infofile);
}

/***********************************************************\
 *  Internal random number generator, since rand()/random() is not
 *  MT-Safe. This RNG is *distinct* from the user RNG, just 
 *  for safety.
 ***********************************************************/
/* code stolen from NetBSD */

/*
 * Cilk_rand_next is stored in the per-processor global section to
 * make this code MT-Safe.
 */
DECLARE_PRIVATE_VARIABLE(static u_long rand_next) = 0;
#ifdef RAND_MAX
#define RTS_RAND_MAX RAND_MAX
#else
#define RTS_RAND_MAX 32767
#endif

static inline int rts_rand(void)
{
     return ((rand_next = rand_next * 1103515245 + 12345)
	     % ((u_int) RTS_RAND_MAX + 1));
}

static inline void rts_srand(u_int seed)
{
     CILK_ASSERT(Cilk_active_size <= RTS_RAND_MAX);
     rand_next = seed;
}

/***********************************************************\
 * Shared memory locks
\***********************************************************/
/*
 *      Lock Diagram:
 * To ensure that no deadlocks occur, make sure that all locks are
 * obtained in order (downward) along a path in the following partial
 * order.  Subcomputation locks must be obtained in a downward order
 * in the subcomputation tree.
 *
 *                      subcomp lock
 *                           |
 *                           |
 *                      stealer lock
 *                      /    |
 *                     /     |
 *      page table lock      |
 *           |         \     |
 *           |          \    |
 *        bs_lock        shmem_lock
 */

/***********************************************************\
 * Cactus-stack memory allocator
\***********************************************************/

/*
 * We should pass this function the frame pointer so we don't have to
 * read it from Cilk_ss->tail.
 */
void *Cilk_alloca(size_t size)
{
     StackFrameT *frame = (StackFrameT *) Cilk_ss->tail;
     void *mem = (char *) frame + frame->size;

     /* align size */
     size = ((size + CILK_MAX_ALIGN - 1) & ~(CILK_MAX_ALIGN - 1));

     CILK_ASSERT((frame->size & (CILK_MAX_ALIGN - 1)) == 0);
     CILK_ASSERT((((int) frame) & (CILK_MAX_ALIGN - 1)) == 0);
     CILK_ASSERT(frame->magic == CILK_FRAME_NONSTEALABLE_MAGIC);

     /* allocate space in frame */
     frame->size = (char *) mem + size - (char *) frame;

     CILK_ASSERT((char *) mem + size
		 <= (char *) Cilk_frame_stack + Cilk_proc_mem);

     return mem;
}

/***********************************************************\
 * Frame stack data structures
\***********************************************************/

/* pointer to StealStuffT array */
DECLARE_SHARED_VARIABLE(static volatile StealStuffT * steal_stuff);


/* 
 * For stealing we use a Dekker-like protocol, that achieves
 * mutual exclusion using shared memory.
 *
 * Thief:
 *   lock
 *   H++     ; signal_intention_to_steal
 *   --MEMBAR-- ensure H write occurs before T is read.
 *   If (H>T)  ; can_steal_p
 *     H--     ; steal_rollback
 *     unlock
 *     give up
 *   steal
 *   unlock
 *
 * Victim
 *   T--
 *   --MEMBAR-- ensure T write occurs before H is read.
 *   if (H>T)
 *      lock
 *      if (H>T)
 *        unlock
 *        give up
 *      unlock
 *
 * The various phases of the thief protocol are abstracted
 * by the following functions.
 */

static inline volatile StackFrameT *signal_intention_to_steal(int pn)
{
     volatile StackFrameT *frame = steal_stuff[pn].head;

     /* 
      * Increment by sizeof(StackFrameT) for now.  We must increment
      * by frame->size eventually.
      */
     steal_stuff[pn].head = frame + 1;
     Cilk_membar_StoreLoad();
     return frame;
}

static inline int can_steal_p(volatile StackFrameT *frame, int pn)
{     
     /* this membar is redundant */
     Cilk_membar_StoreLoad();
     return !(frame + 1 > steal_stuff[pn].tail);
}

static inline void steal_rollback(volatile StackFrameT *frame, int pn)
{
     Cilk_membar_StoreStore();
     steal_stuff[pn].head = frame;
}

/*
 * confirm our intention to steal and make sure we read fresh
 * values from the memory
 */
static inline void confirm_steal(void)
{
     Cilk_membar_LoadLoad();
}

static inline void steal_commit(volatile StackFrameT *next_frame, int pn)
{
     Cilk_membar_StoreStore();
     steal_stuff[pn].head = next_frame;
}

/***********************************************************\
 * Frame stack
\***********************************************************/

static void Cilk_frame_stack_init(void)
{
     Cilk_proc_mem = ROUND_2_PAGE(Cilk_proc_mem);

     CILK_CHECK(PAGE_ALIGNED_P(Cilk_proc_mem),
		("Cilk_proc_mem not multiple of Cilk_pagesize!\n"));

     Cilk_frame_stack = Cilk_map_alloc(Cilk_proc_mem);
     Cilk_last_mapped_page = -1;

     CILK_CHECK(PAGE_ALIGNED_P(Cilk_frame_stack),
	("couldn't align frame_stack to a multiple of Cilk_pagesize\n"));

     Cilk_allow_faults = 0;
     TRACE(TRACE_SETUP, ("frame stack at %x\n", (unsigned int) Cilk_frame_stack));
}

/* Functions to map and unmap stack pages */

#if CILK_DEBUG < 2
/* normal version */
void Cilk_invalidate_stack_pages(int first, int last)
{
     /* invalidate stack pages from first to last, inclusive */
     if (first <= last) {
	  Cilk_map_fault(PAGE_2_STACK_ADDR(first),
			 PAGE_2_BYTES(last - first + 1));
     }
}

static void map_bstore_page(void *addr, int bs_page)
{
     Cilk_map_bstore_page(addr, bs_page);
}

#define vm_check(flag)

#else
/* paranoid version */
#define SHADOW_MAP_SIZE 8192 /* 64M worth of pages */
DECLARE_PRIVATE_VARIABLE(int vm_shadow_map[SHADOW_MAP_SIZE]);

void Cilk_invalidate_stack_pages(int first, int last)
{
     int i;
     /* invalidate stack pages from first to last, inclusive */
     if (first <= last) {
	  Cilk_map_fault(PAGE_2_STACK_ADDR(first),
			 PAGE_2_BYTES((last - first) + 1));
	  
	  for (i = first; i <= last; ++i)
	       vm_shadow_map[i] = -1;
     }
}

static void map_bstore_page(void *addr, int bs_page)
{
     Cilk_map_bstore_page(addr, bs_page);
     vm_shadow_map[STACK_ADDR_2_PAGE(addr)] = bs_page;
}

static void vm_check(int flag)
{
     /* 
      * perform some paranoid check on the consistency of the
      * virtual memory.
      * If flag = 1, initialize the package
      */
     int i;

     if (flag) {
	  for (i = 0; i < SHADOW_MAP_SIZE; ++i)
	       vm_shadow_map[i] = -1;
	  return;
     }

     /*
      * make sure that all pages after Cilk_last_mapped_page
      * are really unmapped.
      */
     for (i = Cilk_last_mapped_page + 1; i < SHADOW_MAP_SIZE; ++i)
	  CILK_ASSERT(vm_shadow_map[i] = -1);

     /*
      * make sure that all pages are mapped where the page
      * table says they should be.
      */
     for (i = 0; i <= Cilk_last_mapped_page; ++i)
	  CILK_ASSERT(vm_shadow_map[i] == -1 ||
		      vm_shadow_map[i] == 
		        Cilk_get_page_table_entry(page_table, i));
}

#endif

/*
 * If fast head-tail check fails, call this slow version.
 */
int Cilk_stolen_check(StackFrameT *frame)
{
     Cilk_sema_wait(&Cilk_ss->stealer_lock);
     if (Cilk_ss->head > frame) {
	  Cilk_sema_signal(&Cilk_ss->stealer_lock);
	  return 1;
     }
     Cilk_sema_signal(&Cilk_ss->stealer_lock);
     return 0;
}

/* 
 * subcomputation creators / destructors 
 */
static volatile SubcomputationT 
*make_subcomputation(SubcomputationT *parent, StackFrameT *frame, int pn)
{
     volatile SubcomputationT *s =
     Cilk_internal_malloc(sizeof(SubcomputationT));
     Cilk_sema_init(&s->lock);
     s->return_subcomp = parent;
     s->frame = frame;
     s->current = pn;
     s->join_count = 0;
     s->return_loc.size = 0;
     s->return_loc.index = 0;
     s->return_loc.inlet = NULL;
     s->return_loc.argsize = 0;
     s->down = s->left = NULL;
     WHEN_CILK_DEBUG(s->inlet_args = NULL);
     WHEN_CILK_FILEIO(s->file_table = NULL);

     /* init measurements */
     s->cp = 0;
     s->work = 0;

     return s;
}

static inline void free_subcomputation(volatile SubcomputationT *s)
{
     Cilk_internal_free(s, sizeof(SubcomputationT));
}

/***********************************************************\
 * page fault catcher
\***********************************************************/

/*
 * Pass this function the fault address for a segmentation violation.
 * Returns 0 if fault was handled successfully, or non-zero if the user
 * program should receive the violation.
 */
int Cilk_signal_handler(void *fault_addr)
{
     signed int fault_page;	/* I emphasise signed -- athena */
     unsigned bs_page;

     Cilk_enter_state(STATE_MAPPING_PAGE);
     /* faults can only happen in user code. Otherwise there is a bug */
     CILK_ASSERT(Cilk_allow_faults);

     if (fault_addr < Cilk_frame_stack ||
	 fault_addr >= Cilk_frame_stack + Cilk_proc_mem) {
	  /* fault did not occur in frame stack - we can't handle it */
	  return 1;
     }
     Cilk_allow_faults = 0;

     fault_page = STACK_ADDR_2_PAGE(fault_addr);
     fault_addr = PAGE_2_STACK_ADDR(fault_page);

     CILK_ASSERT(fault_page >= 0 &&
		 PAGE_2_BYTES(fault_page) < Cilk_proc_mem);

     if (fault_page > Cilk_last_mapped_page)
	  Cilk_last_mapped_page = fault_page;

     bs_page = Cilk_get_page_table_entry(page_table, fault_page);

     TRACE(TRACE_SIGNAL,
     ("mapping stack page %d to bstore page %d\n", fault_page, bs_page));

     CILK_ASSERT(bs_page * Cilk_pagesize + Cilk_pagesize <= Cilk_bstore_mem);

     /* map the just-faulted page to backing store page */
     map_bstore_page(fault_addr, bs_page);
     Cilk_stat_page_fault();

     vm_check(0);
     Cilk_allow_faults = 1;

     Cilk_exit_state(STATE_MAPPING_PAGE);
     return 0;
}

/*
 *  Returns a pointer to the ProcInfoT structure describing entry <entry>
 *  of procedure whose signature is p.
 */
static inline ProcInfoT *get_proc_info(ProcInfoT *p, int entry)
{
     CILK_ASSERT(entry > 0);
     return p + entry;
}

/*
 *  Returns a pointer to the slow version for a procedure
 *  whose signature is p.
 */
static void (*get_proc_slow(ProcInfoT *p)) () {
     return p[0].inlet;
}

/***********************************************************\
 * Initialize system
\***********************************************************/

static void global_init(void)
{
     int i;

     /* set weird self value */
     Self = -1;

     /* get OS page size */
     Cilk_pagesize = Cilk_get_page_size();
     TRACE(TRACE_MISC, ("pagesize = %d\n", Cilk_pagesize));

     /* setup per-processor and global variables */
     Cilk_map_global_variables();

     /* set up malloc() arena */
     Cilk_init_arena();

     /* set up intrnal shared memory heap */
     Cilk_internal_malloc_init();

     /* set up SEGV catcher */
     Cilk_setup_fault_handler();

     /* set up backing store file */
     Cilk_bstore_init();

     /* set up virtual memory for frame stack */
     Cilk_frame_stack_init();

     /* set up frame stack data structures */
     steal_stuff = Cilk_malloc(Cilk_active_size * sizeof(StealStuffT) +
			       CACHE_LINE - 1);
     steal_stuff = NEXT_CACHE_LINE(steal_stuff);
     for (i = 0; i < Cilk_active_size; i++) {
	  Cilk_sema_init(&steal_stuff[i].stealer_lock);
	  Cilk_sema_wait(&steal_stuff[i].stealer_lock);

	  steal_stuff[i].head = Cilk_frame_stack;
	  steal_stuff[i].tail = Cilk_frame_stack;
	  steal_stuff[i].subcomp = NULL;
     }

     /* other shared memory stuff */
     done = 0;

     /* set up page table for each processor */
     Cilk_page_table_init();

     /* initialize the statistics gatherer and timer */
     Cilk_stats_gathering_init();
     Cilk_time_gathering_init();

     Cilk_init_hooks();

     Cilk_run_hooks(Cilk_initial_global_hooks);

     /* fork children */
     if (Cilk_begin_parallel()) {
	  /* everybody is done: print stats and exit */
	  print_all_statistics();

	  /* 
	   * check consistency of the backing store and of the
	   * internal heap, for debugging purposes (catches memory 
	   * leaks in the rts.)
	   */
	  WHEN_CILK_DEBUG(Cilk_check_bstore());
	  WHEN_CILK_DEBUG(Cilk_check_internal_malloc());
	  exit(retval);
     }
}

/***********************************************************\
 * Per-processor initialization
\***********************************************************/

static void scheduler_init(void)
{
     /* 
      * make sure random seed is set to something different in each 
      * process 
      */
     rts_srand(Self + 1);

     /* 
      * Set up Cilk_ss pointer.
      */
     Cilk_ss = &steal_stuff[Self];

     /* initialize the paranoid vm checks */
     vm_check(1);

     Cilk_run_hooks(Cilk_initial_local_hooks);
}

/***********************************************************\
 * Abort code
\***********************************************************/

/*
 * Free child subcomputation structures of <s>.  Each child
 * of <s> is checked to make sure it has returned.  Requires
 * that <s> be locked.  DOES NOT release <s>'s lock before returning.
 */
static void free_children(volatile SubcomputationT *s)
{
     volatile SubcomputationT *child;
     volatile SubcomputationT *next;

     /* splice out child list */
     child = s->down;
     s->down = NULL;

     /* free all children in this list */
     while (child) {
	  WHEN_CILK_DEBUG(Cilk_sema_wait(&child->lock));
	  next = child->left;

	  /* All we need to do is deallocate the child. */
	  CILK_ASSERT(child->state == SUBCOMP_RETURNED);
	  CILK_ASSERT(child->down == NULL);
	  WHEN_CILK_DEBUG(child->state = SUBCOMP_INVALID);
	  WHEN_CILK_DEBUG(Cilk_sema_signal(&child->lock));
	  CILK_ASSERT(child->inlet_args == NULL);
	  free_subcomputation(child);
	  child = next;
     }
}

/*
 * Aborts all children of subcomputation <s>.  Requires
 * that <s> be locked.  DOES NOT release <s>'s lock before returning.
 */
static void abort_children(volatile SubcomputationT *s)
{
     volatile SubcomputationT *child;
     volatile SubcomputationT *next;
     int pn;

     /* splice out child list */
     child = s->down;
     s->down = NULL;

     /* abort all children in this list */
     for (; child; child = next) {
	  TRACE(TRACE_ABORT,
		("abort_children: waiting for child lock, child = %x\n",
		 (int) child));
	  Cilk_sema_wait(&child->lock);
	  TRACE(TRACE_ABORT,
		("abort_children: got child lock\n"));
	  next = child->left;

	  switch (child->state) {
	      case SUBCOMP_RETURNED:
		   TRACE(TRACE_ABORT,
			 ("child is SUBCOMP_RETURNED\n"));
		   /* All we need to do is deallocate the child. */
		   CILK_ASSERT(child->down == NULL);
		   WHEN_CILK_DEBUG(child->state = SUBCOMP_INVALID);
		   Cilk_sema_signal(&child->lock);
		   CILK_ASSERT(child->inlet_args == NULL);
		   free_subcomputation(child);
		   break;

	      case SUBCOMP_SYNCD:
	      case SUBCOMP_SYNCD_FAKE:
		   /* set state to aborting */
		   TRACE(TRACE_ABORT,
			 ("child is SUBCOMP_SYNCD\n"));
		   child->state = SUBCOMP_ABORTING;
		   child->return_loc.size = 0;
		   if (child->return_loc.argsize > 0) {
			CILK_ASSERT(child->inlet_args != NULL);
			Cilk_internal_free(child->inlet_args,
					   child->return_loc.argsize);
			child->return_loc.argsize = 0;
			WHEN_CILK_DEBUG(child->inlet_args = NULL);
		   }
		   if (child->return_loc.inlet) {
			child->return_loc.inlet = NULL;
		   }
		   /* recursively abort its children */
		   abort_children(child);
		   Cilk_sema_signal(&child->lock);
		   break;

	      case SUBCOMP_RUNNING:
		   TRACE(TRACE_ABORT,
			 ("child is SUBCOMP_RUNNING\n"));
		   /* set state to aborting, set void return value */
		   child->state = SUBCOMP_ABORTING;
		   child->return_loc.size = 0;
		   if (child->return_loc.argsize > 0) {
			CILK_ASSERT(child->inlet_args != NULL);
			Cilk_internal_free(child->inlet_args,
					   child->return_loc.argsize);
			child->return_loc.argsize = 0;
			WHEN_CILK_DEBUG(child->inlet_args = NULL);
		   }

		   child->return_loc.inlet = NULL;

		   /* stop running computation */
		   pn = child->current;
		   CILK_ASSERT(pn != -1);
		   Cilk_sema_wait(&steal_stuff[pn].stealer_lock);
		   CILK_ASSERT(steal_stuff[pn].subcomp == child);
		   steal_commit((StackFrameT *) 
				((char *)Cilk_frame_stack + Cilk_proc_mem),
				pn);

		   Cilk_sema_signal(&steal_stuff[pn].stealer_lock);

		   /* recursively abort its children */
		   abort_children(child);
		   Cilk_sema_signal(&child->lock);
		   break;

	      case SUBCOMP_ABORTING:
		   Cilk_die("SUBCOMP_ABORTING state in abort_children!\n");
		   break;

	      default:
		   Cilk_die("bad child state!\n");
		   break;
	  }
     }
     TRACE(TRACE_ABORT, ("end of abort_children\n"));
}

/* subcomp must be locked */
static void abort_parent_subcomputation(volatile SubcomputationT *subcomp)
{
     int pn;
     volatile SubcomputationT *news;
     volatile StackFrameT *frame, *next_frame;

     CILK_ASSERT(subcomp->state == SUBCOMP_RUNNING);

     pn = subcomp->current;
     CILK_ASSERT(pn != -1);
     TRACE(TRACE_ABORT,
	   ("abort_parent_subcomputation: waiting for steal lock\n"));
     Cilk_sema_wait(&steal_stuff[pn].stealer_lock);
     CILK_ASSERT(steal_stuff[pn].subcomp == subcomp);

     /* 
      * we must pretend to steal the top frame, and then immediately sync.
      */
     frame = signal_intention_to_steal(pn);
     if (!can_steal_p(frame, pn)) {
	  /* no child is currently running - we can just do nothing. */
	  steal_rollback(frame, pn);
	  TRACE(TRACE_ABORT,
		("abort_parent_subcomputation: nothing to steal\n"));
	  Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	  return;
     }

     confirm_steal();
     TRACE(TRACE_ABORT,
	   ("abort_parent_subcomputation: steal succeeded\n"));

     /* calculate next frame */
     Cilk_allow_faults = 1;
     next_frame = (StackFrameT *) ((char *) frame + frame->size);
     Cilk_allow_faults = 0;

     /* "victim" gets a new subcomputation */
     news = make_subcomputation((SubcomputationT *) subcomp,
				(StackFrameT *) next_frame, pn);
     news->state = SUBCOMP_ABORTING;
     news->table = Cilk_dup_page_table(subcomp->table);
     WHEN_CILK_DEBUG(
	  news->page_of_frame =
	    Cilk_get_page_table_entry(news->table,
				 STACK_ADDR_2_PAGE(frame)));
     news->parent_index = subcomp->current_index;
     news->current_index = 1;

     WHEN_CILK_FILEIO(Cilk_fileio_steal_hook(news, subcomp));

     CILK_ASSERT_DUMP_SUBCOMP(
	  subcomp->page_of_frame == 	       
	  Cilk_get_page_table_entry(subcomp->table, 
			       STACK_ADDR_2_PAGE(subcomp->frame)),
	  (SubcomputationT *)subcomp);
     CILK_ASSERT_DUMP_SUBCOMP(
	  subcomp->frame == frame,
	  (SubcomputationT *)subcomp);
     CILK_ASSERT_DUMP_SUBCOMP(
	  subcomp->state = SUBCOMP_RUNNING,
	  (SubcomputationT *)subcomp);
     WHEN_CILK_DEBUG(Cilk_allow_faults = 1;
		     subcomp->frame->magic = CILK_FRAME_PROVSTEAL_MAGIC;
		     Cilk_allow_faults = 0);
     subcomp->current = -1;
     subcomp->state = SUBCOMP_SYNCD_FAKE;
     Cilk_membar_StoreStore();
     subcomp->join_count++;
     subcomp->current_index++;
     vm_check(0);

     CILK_ASSERT(steal_stuff[pn].subcomp == subcomp);
     steal_stuff[pn].subcomp = news;

     /* now pretend we abort the newly created subcomputation */
     steal_commit((volatile StackFrameT *) 
		  ((char *)Cilk_frame_stack + Cilk_proc_mem), pn);

     TRACE(TRACE_ABORT,
	   ("abort_parent_subcomputation: created subcomp %x\n", (int) news));

     Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
}

void Cilk_abort_from_slow(void)
{
     volatile SubcomputationT *subcomp;

     TRACE(TRACE_ABORT, ("Cilk_abort_from_slow()\n"));

     /* 
      * this code relies on the fact that head >= tail ==> 
      * subcomp is not changed! 
      */
     CILK_ASSERT(Cilk_ss->head >= Cilk_ss->tail);

     subcomp = Cilk_ss->subcomp;
     TRACE(TRACE_ABORT,
       ("Cilk_abort_from_slow: waiting for subcomp lock %x\n", subcomp));
     Cilk_sema_wait(&subcomp->lock);
     TRACE(TRACE_ABORT,
	   ("Cilk_abort_from_slow: waiting for stealer lock\n"));
     Cilk_sema_wait(&Cilk_ss->stealer_lock);
     TRACE(TRACE_ABORT,
	   ("Cilk_abort_from_slow: got stealer lock\n"));
     CILK_ASSERT(subcomp->current == Self);
     CILK_ASSERT(Cilk_ss->subcomp == subcomp);

     abort_children(subcomp);
     Cilk_sema_signal(&subcomp->lock);
     Cilk_sema_signal(&Cilk_ss->stealer_lock);
}

/***********************************************************\
 * Slow synchronize
\***********************************************************/

int Cilk_sync(StackFrameT *frame)
{
     int jc;
     CILK_ASSERT((char *)frame == (char *)Cilk_ss->tail);
     CILK_ASSERT((char *)Cilk_ss->head >= (char *)Cilk_ss->tail);
     CILK_ASSERT(frame->magic == CILK_FRAME_NONSTEALABLE_MAGIC);
     CILK_ASSERT(Cilk_ss->subcomp != NULL);
     CILK_ASSERT(Cilk_ss->subcomp->current == Self);
     jc = Cilk_ss->subcomp->join_count;
     Cilk_membar_LoadLoad();
     if (jc == 0)
       {
	 free_children(Cilk_ss->subcomp);
	 WHEN_CILK_FILEIO(Cilk_fileio_sync_hook(Cilk_ss->subcomp));
       }
     return jc;
}

/***********************************************************\
 * Return from a subcomputation
\***********************************************************/

/* 
 * apply_result does whatever <info> says to do to <frame>
 * when it returns.
 */
static void apply_result(ProcInfoT *info, void *inlet_args, 
			 volatile SubcomputationT *return_subcomp,
			 StackFrameT *frame, StackFrameT *returning_frame)
{
     int index = info->index;
     int size = info->size;
     int page;

     /* 
      * Install page table of returner.  The old page table and new
      * page table are guaranteed to agree, except that any pages
      * alloca'd by the returned-to procedure since the returning
      * procedure was spawned are not mapped correctly.  We remedy
      * this by invalidating any pages of the returning procedure
      * and then installing the new page table.
      */
     page = STACK_ADDR_2_PAGE(returning_frame);

     if (page <= Cilk_last_mapped_page) {
	  Cilk_invalidate_stack_pages(page, Cilk_last_mapped_page);
	  Cilk_last_mapped_page = page - 1;
     }
     Cilk_free_page_table(page_table);
     page_table = Cilk_dup_page_table(return_subcomp->table);

     Cilk_resume = Cilk_abort = 0;
     Cilk_allow_faults = 1;

     if (info->inlet) {
	  void (*inlet) (void *, void *, void *) = info->inlet;

	  /* add just-returned value to beginning of argument list */
	  memcpy((void *) inlet_args, Cilk_result, size);

	  /* run inlet */
	  inlet(frame, inlet_args, (char *) frame + index);
     } else {
	  /* copy just-returned value into frame */
	  if (size)
	       memcpy((char *) frame + index, Cilk_result, size);
     }
     Cilk_allow_faults = 0;
}

static StackFrameT *suspend_subcomp(volatile SubcomputationT *subcomp)
{
     /* 
      * This code is invoked when a subcomputation suspends without
      * producing a result. This happens when the subcomputation
      * suspended at a sync.
      *
      * subcomp must be locked locked when this function is called.
      * This function releases the lock.
      */
     StackFrameT *return_frame;

     TRACE(TRACE_RETURN_SUBCOMP,
	   ("suspending subcomp %x\n", (int) subcomp));
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("join counter is %d, current = %d return = %x \n",
	    subcomp->join_count,
	    subcomp->current,
	    subcomp->return_subcomp));

     CILK_ASSERT_DUMP_SUBCOMP(
	  subcomp->page_of_frame == 	       
	  Cilk_get_page_table_entry(subcomp->table, 
			       STACK_ADDR_2_PAGE(subcomp->frame)),
	  (SubcomputationT *)subcomp);

     CILK_ASSERT(subcomp->frame != NULL);
     WHEN_CILK_DEBUG(Cilk_allow_faults = 1);
     CILK_ASSERT((subcomp->frame)->magic == CILK_FRAME_NONSTEALABLE_MAGIC);
     WHEN_CILK_DEBUG(Cilk_allow_faults = 0);

     if (subcomp->join_count == 0) {
	  /* 
	   * the join_count was non-zero when the subcomputation
	   * suspended, but it is zero now.  We need to restart
	   * this subcomputation.  We will treat this condition as
	   * a provably good steal.
	   */
	  TRACE(TRACE_RETURN_SUBCOMP,
		("restarting subcomp %x\n", (int) subcomp));

	  /* splice subcomputation back into steal stuff */
	  Cilk_ss->subcomp = subcomp;
	  subcomp->current = Self;

	  return_frame = subcomp->frame;
	  Cilk_sema_signal(&subcomp->lock);
	  return return_frame;
     }

     /* 
      * check that the frame is non stealable, but make it stealable,
      * for use of provably good steals.
      */
     WHEN_CILK_DEBUG(Cilk_allow_faults = 1);
     CILK_ASSERT((subcomp->frame)->magic ==
		 CILK_FRAME_NONSTEALABLE_MAGIC);
     WHEN_CILK_DEBUG((subcomp->frame)->magic = CILK_FRAME_PROVSTEAL_MAGIC);
     WHEN_CILK_DEBUG(Cilk_allow_faults = 0);

     /* set subcomputation state */
     if (subcomp->state != SUBCOMP_ABORTING)
	  subcomp->state = SUBCOMP_SYNCD;

     Cilk_sema_signal(&subcomp->lock);
     return NULL;
}

static void returnit_from_main(volatile SubcomputationT *subcomp)
{
     /* 
      * the suspended subcomputation produces a value returned by
      * cilk_main. subcomp is locked at this point.
      */
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("return of cilk_main\n"));
     retval = *((int *) Cilk_result);

     /* update measures */
     WHEN_CILK_CP(Cilk_critical_path = Cilk_cp_hack);
     WHEN_CILK_CP(Cilk_total_work = Cilk_work_hack);

     done = 1;

     /* finish with returning subcomputation */
     CILK_ASSERT(subcomp->state == SUBCOMP_RUNNING);
     free_children(subcomp);
     Cilk_free_page_table(subcomp->table);
     WHEN_CILK_DEBUG(subcomp->table = NULL);
     WHEN_CILK_DEBUG(subcomp->state = SUBCOMP_INVALID);
     CILK_ASSERT(subcomp->inlet_args == NULL);
     /* To do: make sure all files are closed. KHR */
     Cilk_sema_signal(&subcomp->lock);
     free_subcomputation(subcomp);
     return;
}

/*
 * Sends the return value of the finished subcomputation to wherever it is
 * supposed to go.  Returns a frame to run obtained by a provably good steal,
 * if any.
 *
 * This code is messy, sorry about that.
 */
static StackFrameT *returnit(void)
{
     volatile SubcomputationT *subcomp;
     volatile SubcomputationT *return_subcomp;
     StackFrameT *returning_frame, *return_frame;
     ProcInfoT return_loc;
     int pn;
     void *inlet_args;

     CILK_ASSERT(Cilk_ss->head >= Cilk_ss->tail);
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("returnit(): head = %x tail = %x\n", (int) Cilk_ss->head,
	    (int) Cilk_ss->tail));
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("returnit(): Cilk_result_valid = %d, result=%d\n",
	    Cilk_result_valid, *(int *) Cilk_result));

     /* 
      * Get current subcomputation.  Make sure we get subcomputation
      * lock before stealer lock.  This code relies on the fact
      * that head >= tail ==> subcomp is not changed!
      */
     subcomp = Cilk_ss->subcomp;
     CILK_ASSERT(subcomp != NULL);

     /* get rid of this and the lock in steal()/main() ? */
     Cilk_sema_signal(&Cilk_ss->stealer_lock);
     Cilk_sema_wait(&subcomp->lock);
     Cilk_sema_wait(&Cilk_ss->stealer_lock);

     CILK_ASSERT(subcomp->current == Self);
     CILK_ASSERT(Cilk_ss->subcomp == subcomp);
     CILK_ASSERT(subcomp->state == SUBCOMP_RUNNING ||
		 subcomp->state == SUBCOMP_ABORTING);

     /* run any hooks that need to be run at return time. */
     Cilk_run_hooks(Cilk_return_hooks);
     WHEN_CILK_FILEIO(Cilk_fileio_stop_hook(subcomp));

     /* splice current subcomp out of steal stuff */
     Cilk_ss->subcomp = NULL;
     subcomp->current = -1;

     /* 
      * Treat an abort as a return with a value if the join count is zero
      * and a return without a value if the join count is non-zero.
      */
     if (subcomp->state == SUBCOMP_ABORTING) {
	  CILK_ASSERT(subcomp->return_loc.size == 0);
	  CILK_ASSERT(subcomp->return_loc.inlet == NULL);
	  if (subcomp->join_count > 0) {
	       Cilk_sema_signal(&subcomp->lock);
	       return NULL;
	  }
	  Cilk_result_valid = 1;
     }

     if (!Cilk_result_valid) {
	  return suspend_subcomp(subcomp);
     }

     /* 
      * Otherwise, the frame produces a result. We must send the result to
      * the correct location.
      */
   produces_result:

     /* some debugging stuff */
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("finishing subcomp %x\n", (int) subcomp));
     WHEN_CILK_TRACE(if (subcomp->return_loc.size == 4)
		     TRACE(TRACE_RETURN_SUBCOMP,
			   ("returning %d\n", *((int *) Cilk_result))));

     CILK_ASSERT(subcomp->join_count == 0);

     /* get subcomputation to return to */
     return_subcomp = subcomp->return_subcomp;

     if (return_subcomp == NULL) {
	  /* return value of cilk_main: we're done! */
          WHEN_CILK_FILEIO(Cilk_fileio_free_table(subcomp->file_table));
	  returnit_from_main(subcomp);
	  return NULL;
     }
     WHEN_CILK_FILEIO(Cilk_fileio_return_hook(subcomp, return_subcomp));

     /* get returning frame */
     returning_frame = subcomp->frame;
     CILK_ASSERT(returning_frame != NULL);

     /* get place to put return value */
     return_loc = subcomp->return_loc;
     inlet_args = (void *) subcomp->inlet_args;
     CILK_ASSERT(return_loc.argsize == 0 || inlet_args != NULL);
     subcomp->return_loc.argsize = 0;
     WHEN_CILK_DEBUG(subcomp->inlet_args = NULL);

     /* finish with returning subcomputation */
     Cilk_free_page_table(subcomp->table);
     WHEN_CILK_DEBUG(subcomp->table = NULL);
     if (subcomp->state == SUBCOMP_ABORTING) {
	  WHEN_CILK_DEBUG(subcomp->state = SUBCOMP_INVALID);
	  CILK_ASSERT(subcomp->down == NULL);
	  Cilk_sema_signal(&subcomp->lock);
	  CILK_ASSERT(inlet_args == NULL);
	  free_subcomputation(subcomp);
     } else {
	  subcomp->state = SUBCOMP_RETURNED;
	  free_children(subcomp);
	  Cilk_sema_signal(&subcomp->lock);
     }

     /* get lock on return frame */
   try_again:
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("trying to acquire subcomp lock, return_subcomp = %x\n",
	    return_subcomp));
     Cilk_sema_wait(&return_subcomp->lock);
     return_frame = return_subcomp->frame;
     CILK_ASSERT(return_frame);
     CILK_ASSERT(returning_frame > return_frame);

     /* 
      * take returned value and apply it to the returned-to frame. 
      * Only do it if the returned-to frame isn't aborted.
      */
     if (return_subcomp->state != SUBCOMP_ABORTING) {
	  pn = return_subcomp->current;
	  if (pn == -1) {
	       /* 
	        * just run it - we don't need to worry about local
	        * worker 
	        */
	       apply_result(&return_loc, inlet_args, return_subcomp,
			    return_frame, returning_frame);
	  } else {
	       /* 
	        * Some processor is working on this subcomputation.
	        * Get a lock on that processor, make sure he's still
	        * the worker, get a lock on the return frame, and then
	        * apply the result.
	        */

	       /* lock owning processor */
	       Cilk_sema_wait(&steal_stuff[pn].stealer_lock);
	       CILK_ASSERT(steal_stuff[pn].subcomp == return_subcomp);

	       /* get lock on frame */
	       return_frame = (StackFrameT *)signal_intention_to_steal(pn);
	       CILK_ASSERT(return_frame == return_subcomp->frame);

	       if (!can_steal_p(return_frame, pn)) {
		    steal_rollback(return_frame, pn);
		    Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
		    Cilk_sema_signal(&return_subcomp->lock);
		    WAIT_A_BIT();
		    goto try_again;
	       }

	       apply_result(&return_loc, inlet_args, return_subcomp,
			    return_frame, returning_frame);

	       /* unlock frame */
	       steal_rollback(return_frame, pn);
	       CILK_ASSERT(return_frame == steal_stuff[pn].head);
	       CILK_ASSERT(returning_frame > return_frame);
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	  }

	  /* the inlet may have aborted --- check for that */
	  if (Cilk_abort) {
	       TRACE(TRACE_ABORT,
		     ("inlet set abort flag\n"));
	       if (return_subcomp->state == SUBCOMP_RUNNING)
		    abort_parent_subcomputation(return_subcomp);
	       abort_children(return_subcomp);
	  }
     }

     /* deallocate inlet arguments, if necessary. */
     if (return_loc.argsize > 0)
	  Cilk_internal_free(inlet_args, return_loc.argsize);

     /* 
      * update critical path and work
      */
     WHEN_CILK_CP(if (Cilk_cp_hack > return_subcomp->cp)
		  return_subcomp->cp = Cilk_cp_hack);
     WHEN_CILK_CP(return_subcomp->work += Cilk_work_hack);

     /* 
      * make sure to commit all changes to memory before enabling
      * a remote subcomputation.
      */
     Cilk_membar_StoreStore();

     /* 
      * decrement join counter of returned-to subcomputation, and 
      * do a provably good steal if necessary.
      */
     TRACE(TRACE_RETURN_SUBCOMP,
	   ("enabling subcomp %x, join = %d (before enabling), "
	    "current = %d\n", (int) return_subcomp,
	    return_subcomp->join_count, return_subcomp->current));
     CILK_ASSERT(return_subcomp->state == SUBCOMP_RUNNING ||
		 return_subcomp->state == SUBCOMP_SYNCD ||
		 return_subcomp->state == SUBCOMP_SYNCD_FAKE ||
		 return_subcomp->state == SUBCOMP_ABORTING);
     return_subcomp->join_count--;
     CILK_ASSERT(return_subcomp->join_count >= 0);

     if (return_subcomp->current == -1 &&
	 (return_subcomp->join_count == 0 || Cilk_resume)) {
	  /* that is, if nobody is working on a ready subcomputation */

	  TRACE(TRACE_RETURN_SUBCOMP,
		("adopting subcomp %x\n", (int) return_subcomp));

	  if (return_subcomp->state == SUBCOMP_ABORTING) {
	       /* 
	        * don't restart the subcomputation.  Just pretend it
	        * immediately returns a value.
	        */
	       TRACE(TRACE_ABORT,
		     ("the adopted subcomputation is aborting\n"));
	       subcomp = return_subcomp;
	       goto produces_result;
	  }

	  vm_check(0);
	  CILK_ASSERT(return_frame == return_subcomp->frame);
	  CILK_ASSERT_DUMP_SUBCOMP(
	       return_subcomp->page_of_frame == 	       
	       Cilk_get_page_table_entry(return_subcomp->table, 
				    STACK_ADDR_2_PAGE(return_subcomp->frame)),
	       (SubcomputationT *)return_subcomp);
	  CILK_ASSERT(page_table == return_subcomp->table);
	  CILK_ASSERT_DUMP_SUBCOMP(
	       return_subcomp->state == SUBCOMP_SYNCD || 
	       return_subcomp->state == SUBCOMP_SYNCD_FAKE,
	       (SubcomputationT *)return_subcomp);
	  vm_check(0);
	  WHEN_CILK_DEBUG(Cilk_allow_faults = 1);
	  CILK_ASSERT_DUMP_SUBCOMP(
	       return_frame->magic == CILK_FRAME_PROVSTEAL_MAGIC,
	       (SubcomputationT *)return_subcomp);
	  WHEN_CILK_DEBUG(Cilk_allow_faults = 0);

	  return_subcomp->current = Self;
	  return_subcomp->state = SUBCOMP_RUNNING;
	  Cilk_ss->subcomp = return_subcomp;

	  Cilk_sema_signal(&return_subcomp->lock);

	  return return_frame;
     }
     Cilk_sema_signal(&return_subcomp->lock);
     return NULL;
}

/***********************************************************\
 * Work stealing
\***********************************************************/

static inline int min(int x, int y)
{
     return x < y ? x : y;
}

static inline int max(int x, int y)
{
     return x > y ? x : y;
}

static void runit(volatile StackFrameT *frame)
{
     /* 
      * execute the frame, and keep doing provably good steals while we can 
      */
     while (frame) {
          WHEN_CILK_FILEIO(Cilk_fileio_start_hook(Cilk_ss->subcomp));
          if (Cilk_ss->subcomp->join_count == 0) {
	       free_children(Cilk_ss->subcomp);
	       WHEN_CILK_FILEIO(Cilk_fileio_sync_hook(Cilk_ss->subcomp));
	  }
	  Cilk_ss->head = frame;
	  Cilk_ss->tail = frame;
	  Cilk_result_valid = 0;
	  TRACE(TRACE_STEAL, ("running at entry %d\n", frame->entry));

	  vm_check(0);
	  Cilk_allow_faults = 1;

	  WHEN_CILK_DEBUG(frame->magic = CILK_FRAME_NONSTEALABLE_MAGIC);

	  /* allow stealing */
	  Cilk_sema_signal(&Cilk_ss->stealer_lock);

	  Cilk_enter_state(STATE_WORKING);
	  (get_proc_slow(frame->sig)) (frame);
	  Cilk_exit_state(STATE_WORKING);

	  /* disallow stealing */
	  Cilk_sema_wait(&Cilk_ss->stealer_lock);

	  Cilk_allow_faults = 0;

	  frame = returnit();
	  TRACE(TRACE_RETURN_SUBCOMP,
		("steal: returnit() = %x\n", (int) frame));
     }
}

static void map_frame_into_local_vm(volatile StackFrameT *frame, 
			      volatile struct PageTableT *table)
{
     int begin = STACK_ADDR_2_PAGE((char *) frame);
     int end = STACK_ADDR_2_PAGE((char *) frame + sizeof(StackFrameT) - 1);
     int page;
     int bs_page;
     void *page_addr;

     for (page = begin; page <= end; page++) {
	  bs_page = Cilk_get_page_table_entry(table, page);
	  page_addr = PAGE_2_STACK_ADDR(page);
	  map_bstore_page(page_addr, bs_page);
	  /* consider this a page fault */
	  Cilk_stat_page_fault();
     }
}

/* 
 * unmap my stack only for the portion that differs from the old
 * page table.  Exception: frame has already been mapped, therefore
 * map around it.
 */
static void unmap_my_stack_lazily(volatile struct PageTableT *table,
				  volatile struct PageTableT *old_table,
				  volatile StackFrameT *frame)
{
     int begin = STACK_ADDR_2_PAGE((char *) frame);
     int end = STACK_ADDR_2_PAGE((char *) frame + sizeof(StackFrameT) - 1);
     int start1;
     int end1 = min(begin - 1, Cilk_last_mapped_page);
     int start2 = end + 1;
     int end2 = Cilk_last_mapped_page;

     /* 
      * Make page faults happen on all pages which:
      * 1) differ from our old table (>= first_diff_page),
      * 2) may be mapped (<= Cilk_last_mapped_page), and
      * 3) we haven't mapped correctly yet (begin to 
      * end, inclusive).
      */
     if (old_table != NULL) {
	  start1 = Cilk_page_tables_cmp(old_table, table);
     } else {
	  start1 = 0;
	  CILK_ASSERT(Cilk_last_mapped_page == -1);
     }

     /* 
      * unmap the portion of stack from page to Cilk_last_mapped_page,
      * excluding the chunk from begin to end 
      * we just mapped in.
      */
     Cilk_invalidate_stack_pages(start1, end1);
     Cilk_invalidate_stack_pages(start2, end2);
     Cilk_last_mapped_page = end;
}

static void steal(void)
{
     int pn;
     volatile StackFrameT *frame;
     volatile StackFrameT *next_frame;
     int page;
     volatile SubcomputationT *s, *news;
     ProcInfoT *info;
     volatile struct PageTableT *table, *old_table;

     CILK_ASSERT(Cilk_ss->head >= Cilk_ss->tail);
     Cilk_enter_state(STATE_STEALING);
     TRACE(TRACE_STEAL, ("stealing\n"));

     CILK_ASSERT(Cilk_ss->subcomp == NULL);

     /* Keep trying until we successfully steal. */
     while (!done) {
	  /* pick a random processor, not ourself */
	  pn = ((unsigned) rts_rand()) % Cilk_active_size;
	  if (pn == Self)
	       continue;

	  TRACE(TRACE_STEAL, ("trying to steal from %d\n", pn));
	  Cilk_stat_steal_attempt();

	  /* try to get exclusive stealer lock */
	  if (!Cilk_sema_try(&steal_stuff[pn].stealer_lock)) {
	       TRACE(TRACE_STEAL, ("lock is unavailable\n", pn));
	       continue;
	  }
	  TRACE(TRACE_STEAL, ("got lock on %d\n", pn));

	  TRACE(TRACE_STEAL, ("trying to get current subcomp on %d\n", pn));
	  s = steal_stuff[pn].subcomp;
	  CILK_ASSERT(s != NULL);

	  if (!Cilk_sema_try(&s->lock)) {
	       TRACE(TRACE_STEAL, 
		     ("current subcomp is unavailable on %d\n", pn));
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	       continue;
	  }
	  CILK_ASSERT(s->current == pn);
	  if (s->state == SUBCOMP_ABORTING) {
	       TRACE(TRACE_STEAL, ("current subcomp is aborting on %d\n", pn));
	       Cilk_sema_signal(&s->lock);
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	       continue;
	  }
	  CILK_ASSERT(s->state == SUBCOMP_RUNNING);
	  TRACE(TRACE_STEAL, ("got current subcomp on %d\n", pn));

	  frame = signal_intention_to_steal(pn);

	  if (!can_steal_p(frame, pn)) {
	       /* nothing to steal -- give up */
	       steal_rollback(frame, pn);

	       TRACE(TRACE_STEAL, ("nothing to steal on %d\n", pn));
	       Cilk_sema_signal(&steal_stuff[pn].stealer_lock);
	       Cilk_sema_signal(&s->lock);
	       continue;
	  }

	  TRACE(TRACE_STEAL, ("stealing from %d\n", pn));
	  Cilk_stat_steal();

	  /* ensure we will read fresh values from memory */
	  confirm_steal();

	  /* 
	   * Get page table of victim and mark it so it won't be deallocated.
	   */
	  table = Cilk_dup_page_table(s->table);

	  /* map pages containing frame header */
	  map_frame_into_local_vm(frame, table);

	  /* now we can assert */
	  CILK_ASSERT(frame->magic == CILK_FRAME_STEALABLE_MAGIC);
	  WHEN_CILK_DEBUG(frame->magic = CILK_FRAME_STOLEN_MAGIC);

	  /* 
	   * now that the memory map is set up, we can
	   * find out the size of the frame we stole, and hence where
	   * the beginning of the next frame is.
	   */
	  next_frame = (StackFrameT *) ((char *) frame + frame->size);

	  /* Set head of victim's queue correctly */
	  steal_commit(next_frame, pn);

	  /* 
	   * We now proceed to arrange the subcomputations
	   * to keep the linkage intact. 
	   */
	  TRACE(TRACE_STEAL,
		("stolen frame has entry %d\n", frame->entry));

	  info = get_proc_info(frame->sig, frame->entry);

	  /* victim gets a new subcomputation */
	  news = make_subcomputation((SubcomputationT *) s,
				     (StackFrameT *) next_frame, pn);
	  news->state = SUBCOMP_RUNNING;
	  news->return_loc = *info;
	  if (news->return_loc.index == -1)
	       news->return_loc.index = frame->index;
	  news->down = NULL;
	  news->left = s->down;
	  news->table = table;
	  WHEN_CILK_DEBUG(
	       news->page_of_frame =
	       Cilk_get_page_table_entry(news->table,
				    STACK_ADDR_2_PAGE(next_frame)));
	  news->parent_index = s->current_index;
	  news->current_index = 1;

	  WHEN_CILK_FILEIO(Cilk_fileio_steal_hook(news, s));

	  TRACE(TRACE_STEAL,
		("new subcomp %x, return to %x, index %d, size %d\n",
		 (int) news, (int) news->return_subcomp,
		 news->return_loc.index, news->return_loc.size));

	  /* 
	   * unmap my stack. This is done in a lazy fashion ---
	   * we try to keep the mapping of pages that are common
	   * between the two subcomputations.
	   */
	  old_table = page_table;
	  unmap_my_stack_lazily(table, old_table, frame);

	  /* 
	   * we now copy the page table of the victim.
	   */
	  page = STACK_ADDR_2_PAGE((char *) next_frame - 1);
	  CILK_ASSERT(page >=  STACK_ADDR_2_PAGE((char *) frame +
			 sizeof(StackFrameT) - 1));
	  s->table = Cilk_copy_page_table(table, page);
	  page_table = s->table;

	  /* 
	   * Undo mark to allow victim's table to be deallocated.
	   */
	  Cilk_free_page_table(table);

	  /*
	   * take care of inlets 
	   * This must be done here, after a new page table has been
	   * set up.
	   */
	  if (info->argsize > 0) {
	       CILK_ASSERT(info->inlet != NULL);
	       /* get arguments of inlet out of frame and into subcomp. */
	       news->inlet_args = Cilk_internal_malloc(info->argsize);
	       vm_check(0);
	       Cilk_allow_faults = 1;
	       memcpy((void *) news->inlet_args,
		      (char *) frame + info->argindex, info->argsize);
	       Cilk_allow_faults = 0;
	  } else {
	       WHEN_CILK_DEBUG(news->inlet_args = NULL);
	  }

	  /* make news the current subcomputation of the victim */
	  steal_stuff[pn].subcomp = news;

	  /* tell victim we're done stealing */
	  Cilk_sema_signal(&steal_stuff[pn].stealer_lock);

	  /* 
	   * stealer inherits victim's subcomputation
	   */
	  TRACE(TRACE_STEAL,
		("stealing subcomp: %x, join = %d\n", (int) s, s->join_count));
	  s->current = Self;
	  s->join_count++;
	  s->down = news;
	  s->current_index++;

	  /* 
	   * Undo mark to allow old page table to be deallocated.
	   */
	  if (old_table != NULL)
	       Cilk_free_page_table(old_table);

	  /* 
	   * if stolen frame's end is not page-aligned, add some size to
	   * it to make it page-aligned. 
	   */
	  next_frame = (StackFrameT *) ROUND_2_PAGE((unsigned) next_frame);
	  frame->size = (char *) next_frame - (char *) frame;

	  /* 
	   * Install new subcomputation on local processor.
	   */
	  Cilk_ss->subcomp = s;
	  Cilk_sema_signal(&s->lock);

	  /* run stolen frame */
	  Cilk_exit_state(STATE_STEALING);
	  runit(frame);
	  return;
     }
     Cilk_exit_state(STATE_STEALING);
}

/***********************************************************\
 * Main routine
\***********************************************************/

/* prototype for cilk_main needed here */
int cilk_main(int argc, char *argv[]);

/* Hack: main is defined to be cilk_main; undef it */
#undef main

int main(int argc, char *argv[])
{
     /* start elapsed time measurement */
     start_time = Cilk_get_wall_time();

     /* parse command-line arguments */
     if (Cilk_parse_command_line(&argc, argv))
	  return 0;

     /* initialize global part of system */
     global_init();

     TRACE(TRACE_MISC, ("running\n"));

     /* start gathering timing statistics */
     Cilk_enter_state(STATE_TOTAL);

     /* initialize per-processor parts of system */
     scheduler_init();

     /* start up cilk_main on processor 0 */
     if (Self == 0) {
	  int res;
	  StackFrameT *frame;
	  volatile SubcomputationT *s =
	  make_subcomputation(NULL, Cilk_frame_stack, 0);
	  s->state = SUBCOMP_RUNNING;
	  s->table = Cilk_create_empty_page_table();
	  Cilk_ss->subcomp = s;
	  page_table = Cilk_dup_page_table(s->table);
	  WHEN_CILK_DEBUG(
	       s->page_of_frame =
	       Cilk_get_page_table_entry(s->table,
				    STACK_ADDR_2_PAGE(Cilk_frame_stack)));
	  s->current_index = 1;
	  s->parent_index = -1;
	  WHEN_CILK_FILEIO(s->file_table = Cilk_fileio_empty_table());

	  WHEN_CILK_FILEIO(Cilk_fileio_start_hook(s));

	  Cilk_result_valid = 0;

	  TRACE(TRACE_MAIN, ("invoking cilk_main()\n"));

	  /* start with a 0 critical path */
	  WHEN_CILK_CP(Cilk_cp_hack = 0);
	  WHEN_CILK_CP(Cilk_work_hack = 0);
	  
	  vm_check(0);
	  Cilk_allow_faults = 1;

	  /* allow stealing */
	  Cilk_sema_signal(&Cilk_ss->stealer_lock);

	  Cilk_enter_state(STATE_WORKING);
	  res = cilk_main(argc, argv);
	  Cilk_exit_state(STATE_WORKING);

	  /* disallow stealing */
	  Cilk_sema_wait(&Cilk_ss->stealer_lock);

	  Cilk_allow_faults = 0;

	  TRACE(TRACE_MAIN, ("cilk_main returned\n"));

	  if (!Cilk_result_valid) {
	       /* cilk_main wasn't stolen - set result values */
	       TRACE(TRACE_MAIN,
		     ("main was not stolen\n"));
	       CILK_ASSERT(Cilk_ss->subcomp->return_subcomp == NULL);
	       *((int *) Cilk_result) = res;
	       Cilk_result_valid = 1;
	  } else {
	       TRACE(TRACE_MAIN,
		     ("main was stolen\n"));
	  }

	  TRACE(TRACE_MAIN, ("main: invoking returnit()\n"));
	  frame = returnit();
	  TRACE(TRACE_RETURN_SUBCOMP,
		("main: returnit() = %x\n", (int) frame));

	  /* keep doing provably good steals while we can */
	  TRACE(TRACE_MAIN, ("main: doing provably good steals\n"));
	  runit(frame);
     }
     TRACE(TRACE_MAIN, ("main: stealing normally\n"));
     /* keep stealing until we're done */
     while (!done)
	  steal();

     /* free any leftover page table */
     if (page_table != NULL)
	  Cilk_free_page_table(page_table);

     TRACE(TRACE_MAIN, ("main: done!\n"));

     /* stop gathering time statistics */
     Cilk_exit_state(STATE_TOTAL);

     /* kill child processes */
     Cilk_end_parallel();

     /* never reached! */
     CILK_ASSERT(0 /* can't happen */ );
     return 0;
}
