/*
 * $Id: debug.c,v 1.3 1996/09/05 19:21:53 randall Exp $
 */

/*
 * $Log: debug.c,v $
 * Revision 1.3  1996/09/05 19:21:53  randall
 * Small changes.
 *
 * Revision 1.2  1996/09/03 19:05:36  athena
 * Added many paranoid tests.  Using munmap() instead of mprotect()
 *
 * Revision 1.1  1996/09/03 14:59:08  athena
 * Added more debugging stuff
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/* this file contains some simple debugging support */
#include <cilk-internal.h>

#if CILK_DEBUG
static char *state_desc[SUBCOMP_MAX_VALID_STATE+1] = {
     "INVALID",
     "RUNNING",
     "SYNCD",
     "ABORTING",
     "RETURNED",
     "SYNCD_FAKE"
};

void Cilk_dump_frame(StackFrameT *f)
{
     extern volatile int Cilk_last_mapped_page;
     extern volatile int Cilk_allow_faults;
     Cilk_allow_faults = 1;
     Cilk_dprintf("Dump of frame %lx:\n", (long) f);
     Cilk_dprintf("size=%d  entry=%d  index=%d  magic=%x\n",
		  f->size,  f->entry, f->index, f->magic);
#if CILK_DEBUG <= 1
     /* this destroys the shadow vm map, therefore better avoid it */
     Cilk_invalidate_stack_pages(0, Cilk_last_mapped_page);
     Cilk_last_mapped_page = -1;
     Cilk_dprintf("New dump of frame %lx after remapping:\n", (long) f);
     Cilk_dprintf("size=%d  entry=%d  index=%d  magic=%x\n",
		  f->size,  f->entry, f->index, f->magic);
     Cilk_allow_faults = 0;
#endif
}

void Cilk_dump_subcomputation(SubcomputationT *s)
{
     Cilk_dprintf("Subcomputation dump for %lx:\n", (long)s);
     Cilk_dprintf("subcomputation is %s\n", 
		  s->lock ? "locked" : "not locked");
     if (s->state >= 0 && s->state <= SUBCOMP_MAX_VALID_STATE) 
	  Cilk_dprintf("state is %s\n", state_desc[s->state]);
     else
	  Cilk_dprintf("state is unknown state %d\n", s->state);

     Cilk_dprintf("stack frame is %lx\n", (long)s->frame);
     Cilk_dprintf("current is %d,   join count is %d\n", (long)s->current,
		  s->join_count);

     Cilk_dump_frame(s->frame);
}

#endif /* CILK_DEBUG */
