/*
 * $Id: hooks.c,v 1.5 1996/09/09 23:38:55 randall Exp $
 */

/*
 * $Log: hooks.c,v $
 * Revision 1.5  1996/09/09 23:38:55  randall
 * Makefile update, some fileio stuff.
 *
 * Revision 1.4  1996/08/27 23:11:48  randall
 * Made the file-pointer-less files operational.
 *
 * Revision 1.3  1996/08/27 19:28:29  randall
 * One pass of cleaning up.
 *
 * Revision 1.2  1996/08/27 15:43:30  athena
 * Some more cleanup -- changed some low-level interfaces.
 *
 * Revision 1.1  1996/08/26 22:16:56  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/* hook manager */
#include <cilk-internal.h>

void Cilk_add_hook(HookList **listp, HookT fn)
{
     HookList *new;

     CILK_ASSERT(Self == -1);
     new = (HookList *)malloc(sizeof (HookList));

     new->fn = fn;
     new->next = NULL_HOOK;

     /* traverse the list */
     while (*listp != NULL_HOOK)
	  listp = &((*listp)->next);

     /* append at the end */
     *listp = new;
}

void Cilk_run_hooks(HookList *list)
{
     for ( ; list != NULL_HOOK ; list = list->next) 
	  list->fn();
}
	  
void Cilk_init_hooks(void)
{
     
}

/* declaration of the various hooks */
HookList *Cilk_initial_global_hooks = NULL_HOOK;
HookList *Cilk_initial_local_hooks = NULL_HOOK;
HookList *Cilk_return_hooks = NULL_HOOK;

#if CILK_FILEIO
struct FileTableT *Cilk_fileio_empty_table(void) __attribute__((weak));
struct FileTableT *Cilk_fileio_empty_table(void){return NULL;}
void Cilk_fileio_free_table(struct FileTableT *table) __attribute__((weak));
void Cilk_fileio_free_table(struct FileTableT *table){}

void Cilk_fileio_start_hook(volatile SubcomputationT *subcomp) __attribute__((weak));
void Cilk_fileio_start_hook(volatile SubcomputationT *subcomp){}
void Cilk_fileio_stop_hook(volatile SubcomputationT *subcomp) __attribute__((weak));
void Cilk_fileio_stop_hook(volatile SubcomputationT *subcomp){}
void Cilk_fileio_steal_hook(volatile SubcomputationT *child,
			    volatile SubcomputationT *parent) __attribute__((weak));
void Cilk_fileio_steal_hook(volatile SubcomputationT *child,
			    volatile SubcomputationT *parent){}
void Cilk_fileio_sync_hook(volatile SubcomputationT *subcomp) __attribute__((weak));
void Cilk_fileio_sync_hook(volatile SubcomputationT *subcomp){}
void Cilk_fileio_return_hook(volatile SubcomputationT *child,
			     volatile SubcomputationT *parent) __attribute__((weak));
void Cilk_fileio_return_hook(volatile SubcomputationT *child,
			     volatile SubcomputationT *parent){}
#endif
