/*
 * $Id: stats.c,v 1.3 1996/09/05 14:51:31 athena Exp $
 */

/*
 * $Log: stats.c,v $
 * Revision 1.3  1996/09/05 14:51:31  athena
 * Rationalization of timers
 *
 * Revision 1.2  1996/08/27 19:28:33  randall
 * One pass of cleaning up.
 *
 * Revision 1.1  1996/08/26 22:17:00  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************\
 * Statistics gathering
\***********************************************************/

#include <cilk-internal.h>

#if CILK_STATS
#include <stdio.h>
extern FILE *Cilk_infofile;

DECLARE_SHARED_VARIABLE(static StatsT * stat_array);

void Cilk_stats_gathering_init(void)
{
     int i;

     stat_array = (StatsT *) Cilk_malloc(sizeof(StatsT) * Cilk_active_size);

     for (i = 0; i < Cilk_active_size; i++) {
	  stat_array[i].steal_attempts = 0;
	  stat_array[i].steals = 0;
	  stat_array[i].page_faults = 0;
	  stat_array[i].nthreads = 0;
	  stat_array[i].null_threads = 0;
	  stat_array[i].broken_threads = 0;
     }
}

void Cilk_print_rts_statistics(void)
{
     int i;
     StatsT total;

     total.steal_attempts = 0;
     total.steals = 0;
     total.page_faults = 0;
     total.nthreads = 0;
     total.null_threads = 0;
     total.broken_threads = 0;

     fprintf(Cilk_infofile,
	     "\nRUNTIME SYSTEM STATISTICS:\n"
	     "\n"
	     " PN   Attempts     Steals     Faults    Threads     Broken       Null\n"
	     "---------------------------------------------------------------------\n");

     for (i = 0; i < Cilk_active_size; i++) {
	  fprintf(Cilk_infofile, "%3d%11d%11d%11d%11ld%11ld%11ld\n", i,
		  stat_array[i].steal_attempts,
		  stat_array[i].steals,
		  stat_array[i].page_faults,
		  stat_array[i].nthreads,
		  stat_array[i].broken_threads,
		  stat_array[i].null_threads);
	  total.steal_attempts += stat_array[i].steal_attempts;
	  total.steals += stat_array[i].steals;
	  total.page_faults += stat_array[i].page_faults;
	  total.nthreads += stat_array[i].nthreads;
	  total.broken_threads += stat_array[i].broken_threads;
	  total.null_threads += stat_array[i].null_threads;
     }

     fprintf(Cilk_infofile,
	     "---------------------------------------------------------------------\n");
     fprintf(Cilk_infofile, "TOT%11d%11d%11d%11ld%11ld%11ld\n", 
	     total.steal_attempts,
	     total.steals,
	     total.page_faults,
	     total.nthreads,
	     total.broken_threads,
	     total.null_threads);

     fprintf(Cilk_infofile, "\n");
}

void Cilk_stat_steal_attempt(void)
{
     stat_array[Self].steal_attempts++;
}

void Cilk_stat_steal(void)
{
     stat_array[Self].steals++;
}

void Cilk_stat_page_fault(void)
{
     stat_array[Self].page_faults++;
}

void Cilk_stat_broken_thread(void)
{
     stat_array[Self].broken_threads++;
}

void Cilk_stat_new_thread(void)
{
     stat_array[Self].nthreads++;
}

void Cilk_stat_null_thread(void)
{
     stat_array[Self].null_threads++;
}

#endif				/* CILK_STATS */
