
/*
 * $Id: watch.c,v 1.3 1996/08/26 22:17:02 athena Exp $
 */
/*
 * $Log: watch.c,v $
 * Revision 1.3  1996/08/26 22:17:02  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 * Revision 1.2  1996/06/07 17:55:51  athena
 * Fixed copyrights. I hate this.
 *
 * Revision 1.1  1996/04/03 20:43:05  randall
 * Added watch utility.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * A neat little utility I wrote to watch a cilk process change states.
 * The states are OS states, not Cilk internal states.  After a Cilk
 * job (or any other job) is running, type 'watch <pid>' where <pid>
 * is the process number obtained using 'ps'.  Watch the OS suspend
 * and resume the process and move it between processors.
 *
 *                                     -Keith
 */

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>

int main(int argc, char *argv[])
{
     pid_t pid;
     int fd;
     char buf[100];
     int init = 0;
     int retval1, retval2;
     hrtime_t t1, t2;

     prstatus_t p;
     int proc, oldproc;
     long flags, oldflags;

     prpsinfo_t pr;
     char sname, oldsname;

     assert(argc == 2);
     pid = atoi(argv[1]);

     sprintf(buf, "/proc/%d", pid);
     fd = open(buf, O_RDONLY);
     if (fd < 0) {
	  printf("bad fd\n");
	  return 0;
     }
     while (1) {
//      t1 = gethrtime();
	  retval1 = ioctl(fd, PIOCPSINFO, &pr);		/* ~65usec */
	  retval2 = ioctl(fd, PIOCSTATUS, &p);	/* ~32usec */
//      t2 = gethrtime();
	  // printf("call: %f\n", (t2 - t1)/1e3);
	  if (retval1 < 0 || retval2 < 0) {
	       if (errno == ENOENT)
		    printf("process terminated\n");
	       else
		    printf("bad retval\n");
	       return 0;
	  }
	  flags = p.pr_flags;
	  proc = p.pr_processor;
	  sname = pr.pr_sname;

	  if (!init) {
	       init = 1;
	       goto next;
	  }
	  if (oldflags != flags)
	       printf("flag change: %8x => %8x\n", oldflags, flags);
	  if (oldproc != proc)
	       printf("moved from processor %d to processor %d\n", oldproc, proc);
	  if (oldsname != sname)
	       printf("changed from state %c to %c\n", oldsname, sname);

	next:
	  oldsname = sname;
	  oldflags = flags;
	  oldproc = proc;
     }
}
